# ifndef CppADCppAD_vectorIncluded
# define CppADCppAD_vectorIncluded

// BEGIN SHORT COPYRIGHT
/* -----------------------------------------------------------------------
CppAD: C++ Algorithmic Differentiation: Copyright (C) 2003-05 Bradley M. Bell

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
------------------------------------------------------------------------ */
// END SHORT COPYRIGHT

/*
$begin CppAD_vector$$
$spell
	Bool
	resize
	cout
	endl
	std
	Cpp
	const
	vec
	ostream
	elem
$$

$index vector, CppAD template class$$
$index class, template CppAD vector$$
$index template, CppAD vector class$$

$section The CppAD::vector Template Class$$

$table 
$bold Syntax$$ $cnext
$code # include <CppAD/CppAD_vector.h>$$ 
$tend

$fend 25$$


$head Description$$
The include file $code CppAD/CppAD_vector.h$$ defines the
vector template class $code CppAD::vector$$.
This is a $xref/SimpleVector/$$ template class and in addition
it has the features listed below:

$head Include$$
The file $code CppAD/CppAD_vector.h$$ is included by $code CppAD/CppAD.h$$
but it can also be included separately with out the rest of the 
CppAD include files.

$head Assignment$$
$index assignment, CppAD vector$$
If $italic x$$ and $italic y$$ are 
$syntax%CppAD::vector<%Type%>%$$ objects,
$syntax%
	%y% = %x%
%$$
has all the properties listed for a
$xref/SimpleVector/Assignment/simple vector assignment/$$
plus the following:
$pre

$$
The $code CppAD::vector$$ template class will check that
the size of $italic x$$ is equal to the size of $italic y$$
before doing the assignment.
If the sizes are not equal, $code CppAD::vector$$ will use
$xref/CppADError/Description/CppADUsageError/CppADUsageError/1/$$
to generate an appropriate error report.
$pre

$$
A reference to the vector $italic y$$ is returned.
An example use of this reference is in multiple assignments of the form
$syntax%
	%z% = %y% = %x%
%$$

$head Element Access$$
$index [], CppAD vector$$
$index vector, [] CppAD$$
If $italic x$$ is a $syntax%CppAD::vector<%Type%>%$$ object
and $code i$$ has type $code size_t$$,
$syntax%
	%x%[%i%]
%$$
has all the properties listed for a
$xref/SimpleVector/Element Access/simple vector element access/$$
plus the following:
$pre

$$
If $italic i$$ is not less than the size of the $italic x$$,
$code CppAD::vector$$ will use
$xref/CppADError/Description/CppADUsageError/CppADUsageError/1/$$
to generate an appropriate error report.

$head Push Back$$
$index push, CppAD vector$$
$index vector, CppAD push back$$
If $italic x$$ is a $syntax%CppAD::vector<%Type%>%$$ object
with size equal to $italic n$$ and
$italic t$$ has type $italic Type$$,
$syntax%
	%x%.push_back(%t%)
%$$
extends the vector $italic x$$ so that its new size is $italic n$$ plus one
and $syntax%%x%[%n%]%$$ is equal to $italic t$$
(in the sense of the $italic Type$$ assignment operator).

$head Output$$
If $italic x$$ is a $syntax%CppAD::vector<%Type%>%$$ object
and $italic os$$ is an $code std::ostream$$,
and the operation
$syntax%
	%os% << %x%
%$$
will output the vector $italic x$$ to the standard
output stream $italic os$$.
The elements of $italic x$$ are enclosed at the beginning by a
$code {$$ character,
they are separated by $code ,$$ characters,
and they are enclosed at the end by $code }$$ character.
It is assumed by this operation that if $italic e$$
is an object with type $italic Type$$,
$syntax%
	%os% << %e%
%$$
will output the value $italic e$$ to the standard
output stream $italic os$$.


$head vectorBool$$
$index vectorBool$$
The file $code <CppAD/CppAD_vector.h>$$ also defines the class
$code CppAD::vectorBool$$.
This has the same specifications as $code CppAD::vector<bool>$$ except
that $code vectorBool$$ conserves on memory
(on the other hand, $code vectorBool$$ is expected to be faster
than $code vectorBool$$).
Another exception is the output operator for $code vectorBool$$
which is specified below:

$subhead Output$$
The $code CppAD::vectorBool$$ output operator
prints each boolean value as 
a $code 0$$ for false,
a $code 1$$ for true, 
and does not print any other output; i.e.,
the vector is written a long sequence of zeros and ones with no
surrounding $code {$$, $code }$$ and with no separating commas or spaces. 

$head Example$$
$children%
	Example/CppAD_vector.cpp%
	Example/vectorBool.cpp
%$$
The files
$xref/CppAD_vector.cpp/$$ and
$xref/vectorBool.cpp/$$ each
contain an example and test of this template class.
They return true if they succeed and false otherwise.

$head Exercise$$
$index exercise, CppAD::vector$$
Create and run a program that contains the following code:
$codep
	CppAD::vector<double> x(3);
	size_t i;
	for(i = 0; i < 3; i++)
		x[i] = 4. - i;
	std::cout << "x = " << x << std::endl;
$$

$end


$end

------------------------------------------------------------------------ 
*/

# include <cstddef>
# include <iostream>
# include <limits>
# include <CppAD/CppADError.h>

# ifndef CppADNull
# define CppADNull 0
# endif

namespace CppAD { //  BEGIN CppAD namespace

// ------------------ CppAD::vector<Type> ----------------------------------

# ifdef NDEBUG
# define CppADvectorAllocate(data, Type, capacity)                         \
	data = new Type[capacity];
# else
# define CppADvectorAllocate(data, Type, capacity)                         \
	{	try                                                        \
		{	data = new Type[capacity];                         \
		}                                                          \
		catch (...)                                                \
		{	CppADUsageError(0,                                 \
			"CppAD::vector: cannot allocate enough memory"     \
			);	                                           \
		}                                                          \
	}
# endif


template <class Type>
class vector {
private:
	size_t capacity;
	size_t length;
	Type   * data;
public:
	// type of the elements in the vector
	typedef Type value_type;

	// default constructor
	inline vector(void) : capacity(0), length(0) , data(CppADNull)
	{ }
	// constructor with a specified size
	inline vector(size_t n) : capacity(n), length(n)
	{	if( length == 0 )
			data = CppADNull;
		else	CppADvectorAllocate(data, Type, capacity);
	}
	// copy constructor
	inline vector(const vector &x) : capacity(x.length), length(x.length)
	{	size_t i;
		if( length == 0 )
			data = CppADNull;
		else	CppADvectorAllocate(data, Type, capacity);

		for(i = 0; i < length; i++)
			data[i] = x.data[i];
	}
	// destructor
	~vector(void)
	{	delete [] data; }

	// size function
	inline size_t size(void) const
	{	return length; }

	// resize function
	inline void resize(size_t n)
	{	length = n;
		if( capacity >= n )
			return;
		if( capacity > 0 )
			delete [] data;
		capacity = n;
		if( capacity == 0 )
			data = CppADNull;
		else	CppADvectorAllocate(data, Type, capacity);
	}
	// assignment operator
	inline vector & operator=(const vector &x)
	{	size_t i;
		CppADUsageError(
			length == x.length ,
			"size miss match in assignment operation"
		);
		for(i = 0; i < length; i++)
			data[i] = x.data[i];
		return *this;
	}
	// non-constant element access
	Type & operator[](size_t i)
	{	CppADUsageError(
			i < length,
			"vector index greater than or equal vector size"
		);
		return data[i]; 
	}
	// constant element access
	const Type & operator[](size_t i) const
	{	CppADUsageError(
			i < length,
			"vector index greater than or equal vector size"
		);
		return data[i]; 
	}
	// add to the back of the array
	void push_back(const Type &x)
	{	CppADUnknownError( length <= capacity );
		if( length == capacity )
		{	// allocate more capacity
			Type *tmp;
			if( capacity == 0 )
				capacity = 2;
			else	capacity = 2 * length;

			CppADvectorAllocate(tmp, Type, capacity);

			size_t i;
			for(i = 0; i < length; i++)
				tmp[i] = data[i];
			delete [] data;
			data = tmp;
		}
		data[length++] = x;
	}
};

// output operator
template <class Type>
inline std::ostream& operator << (
	std::ostream              &os  , 
	const CppAD::vector<Type> &vec )
{	size_t i = 0;
	size_t n = vec.size();

	os << "{ ";
	while(i < n)
	{	os << vec[i++]; 
		if( i < n )
			os << ", ";
	}
	os << " }";
	return os;
}

/*
--------------------------- vectorBool -------------------------------------
*/
# undef  CppADvectorAllocate
# ifdef NDEBUG
# define CppADvectorAllocate(data, UnitType, nunit)                        \
	data = new UnitType[nunit]; 
# else
# define CppADvectorAllocate(data, UnitType, nunit)                        \
	{	try                                                        \
		{	data = new UnitType[nunit];                        \
		}                                                          \
		catch (...)                                                \
		{	CppADUsageError(0,                                 \
			"CppAD::vectorBool: cannot allocate enough memory" \
			);	                                           \
		}                                                          \
	}
# endif

class vectorBoolElement {
	typedef size_t UnitType;
private:
	UnitType *unit;
	UnitType mask;
public:
	vectorBoolElement(UnitType *unit_, UnitType mask_)
	: unit(unit_) , mask(mask_)
	{ }
	vectorBoolElement(const vectorBoolElement &e)
	: unit(e.unit) , mask(e.mask)
	{ }
	operator bool() const
	{	return (*unit & mask) != 0; }
	vectorBoolElement& operator=(bool bit)
	{	if(bit)
			*unit |= mask;
		else	*unit &= ~mask;
		return *this;
	} 
};

class vectorBool {
	typedef size_t UnitType;
private:
	static const  size_t BitPerUnit 
		= std::numeric_limits<UnitType>::digits;
	size_t    nunit;
	size_t    length;
	UnitType *data;
public:
	// type of the elements in the vector
	typedef bool value_type;

	// default constructor
	inline vectorBool(void) : nunit(0), length(0) , data(CppADNull)
	{ }
	// constructor with a specified size
	inline vectorBool(size_t n) : nunit(0), length(0), data(CppADNull)
	{	if( n != 0 )
		{	nunit    = (n - 1) / BitPerUnit + 1;
			length   = n;

			CppADvectorAllocate(data, UnitType, nunit);
		}
	}
	// copy constructor
	inline vectorBool(const vectorBool &v) 
	: nunit(v.nunit), length(v.length)
	{	size_t i;
		if( nunit == 0 )
			data = CppADNull;
		else	CppADvectorAllocate(data, UnitType, nunit);

		for(i = 0; i < nunit; i++)
			data[i] = v.data[i];
	}
	// destructor
	~vectorBool(void)
	{	delete [] data; }

	// size function
	inline size_t size(void) const
	{	return length; }

	// resize function
	inline void resize(size_t n)
	{	length = n;
		if( nunit * BitPerUnit >= n )
			return;
		if( nunit > 0 )
			delete [] data;
		if( n == 0 )
			data = CppADNull;
		else
		{	nunit    = (n - 1) / BitPerUnit + 1;
			CppADvectorAllocate(data, UnitType, nunit);
		}
	}
	// assignment operator
	inline vectorBool & operator=(const vectorBool &v)
	{	size_t i;
		CppADUsageError(
			length == v.length ,
			"size miss match in assignment operation"
		);
		CppADUnknownError( nunit == v.nunit );
		for(i = 0; i < nunit; i++)
			data[i] = v.data[i];
		return *this;
	}
	// non-constant element access
	vectorBoolElement operator[](size_t k)
	{	size_t i, j;
		CppADUsageError(
			k < length,
			"vector index greater than or equal vector size"
		);
		i    = k / BitPerUnit;
		j    = k - i * BitPerUnit;
		return vectorBoolElement(data + i , UnitType(1) << j );
	}
	// constant element access
	bool operator[](size_t k) const
	{	size_t i, j;
		UnitType unit;
		UnitType mask;
		CppADUsageError(
			k < length,
			"vector index greater than or equal vector size"
		);
		i    = k / BitPerUnit;
		j    = k - i * BitPerUnit;
		unit = data[i];
		mask = UnitType(1) << j;
		return (unit & mask) != 0;
	}
	// add to the back of the array
	void push_back(bool bit)
	{	size_t i, j;
		UnitType mask;
		CppADUnknownError( length <= nunit * BitPerUnit );
		if( length == nunit * BitPerUnit )
		{	// allocate another unit
			UnitType *tmp;
			nunit++;

			CppADvectorAllocate(tmp, UnitType, nunit);

			for(i = 0; i < nunit; i++)
				tmp[i] = data[i];
			delete [] data;
			data = tmp;
		}
		i    = length / BitPerUnit;
		j    = length - i * BitPerUnit;
		mask = UnitType(1) << j;
		if( bit )
			data[i] |= mask;
		else	data[i] &= ~mask;
		length++;
	}
};

// output operator
inline std::ostream& operator << (
	std::ostream     &os  , 
	const vectorBool &v   )
{	size_t i = 0;
	size_t n = v.size();

	while(i < n)
		os << v[i++]; 
	return os;
}

# undef  CppADvectorAllocate

} // END CppAD namespace


# endif
