/* 
 * MBDyn (C) is a multibody analysis code. 
 * http://www.mbdyn.org
 *
 * Copyright (C) 1996-2005
 *
 * Pierangelo Masarati	<masarati@aero.polimi.it>
 * Paolo Mantegazza	<mantegazza@aero.polimi.it>
 *
 * Dipartimento di Ingegneria Aerospaziale - Politecnico di Milano
 * via La Masa, 34 - 20156 Milano, Italy
 * http://www.aero.polimi.it
 *
 * Changing this copyright notice is forbidden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation (version 2 of the License).
 * 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* vettori 3 e matrici 3x3 */

#ifdef HAVE_CONFIG_H
#include <mbconfig.h>           /* This goes first in every *.c,*.cc file */
#endif /* HAVE_CONFIG_H */

#include <cmath>
#include <ac/float.h>

#include <matvec3.h>

/* noteworthy constant */
const Mat3x3 Eye3(1., 0., 0., 0., 1., 0., 0., 0., 1.);
const Mat3x3 Zero3x3(0., 0., 0., 0., 0., 0., 0., 0., 0.);
const Vec3 Zero3(0., 0., 0.);


/* Vec3 - begin */
template<class T>
Mat3x3T<T> Vec3T<T>::Tens(const Vec3T<T>& v) const
{
   return Mat3x3T<T>(pdVec[V1]*v.pdVec[V1],
		 pdVec[V2]*v.pdVec[V1],
		 pdVec[V3]*v.pdVec[V1],
		 pdVec[V1]*v.pdVec[V2],
		 pdVec[V2]*v.pdVec[V2],
		 pdVec[V3]*v.pdVec[V2],
		 pdVec[V1]*v.pdVec[V3],
		 pdVec[V2]*v.pdVec[V3],
		 pdVec[V3]*v.pdVec[V3]);
}

/* Prodotto "tensore".  Restituisce se stesso per se stesso */
template<class T>
Mat3x3T<T> Vec3T<T>::Tens(void) const
{
	return Tens(*this);
}     

/* Prodotto vettore per matrice */
template<class T>
Mat3x3T<T> Vec3T<T>::Cross(const Mat3x3T<T>& m) const {
   return Mat3x3T<T>(pdVec[V2]*m.pdMat[M31]-pdVec[V3]*m.pdMat[M21],
		 pdVec[V3]*m.pdMat[M11]-pdVec[V1]*m.pdMat[M31],
		 pdVec[V1]*m.pdMat[M21]-pdVec[V2]*m.pdMat[M11],
		 pdVec[V2]*m.pdMat[M32]-pdVec[V3]*m.pdMat[M22],
		 pdVec[V3]*m.pdMat[M12]-pdVec[V1]*m.pdMat[M32],
		 pdVec[V1]*m.pdMat[M22]-pdVec[V2]*m.pdMat[M12],
		 pdVec[V2]*m.pdMat[M33]-pdVec[V3]*m.pdMat[M23],
		 pdVec[V3]*m.pdMat[M13]-pdVec[V1]*m.pdMat[M33],
		 pdVec[V1]*m.pdMat[M23]-pdVec[V2]*m.pdMat[M13]);
}

/* Vec3T - end */

/* Mat3x3T - begin */

/* inversione */
template<class T>
T Mat3x3T<T>::dDet(void) const
{
   T* p = (T*)pdMat;

   return p[M11]*(p[M22]*p[M33]-p[M23]*p[M32])
     +p[M12]*(p[M23]*p[M31]-p[M21]*p[M33])
     +p[M13]*(p[M21]*p[M32]-p[M22]*p[M31]);
}
   
/* inversione */
template<class T>
Mat3x3T<T> Mat3x3T<T>::Inv(const T &d) const
{
   ASSERT(fabs(d) > DBL_EPSILON);

   T* p = (T*)pdMat;

   return Mat3x3T((p[M22]*p[M33]-p[M23]*p[M32])/d,
		 (p[M23]*p[M31]-p[M21]*p[M33])/d,
		 (p[M21]*p[M32]-p[M22]*p[M31])/d,
		 (p[M13]*p[M32]-p[M12]*p[M33])/d,
		 (p[M11]*p[M33]-p[M13]*p[M31])/d,
		 (p[M12]*p[M31]-p[M11]*p[M32])/d,
		 (p[M12]*p[M23]-p[M13]*p[M22])/d,
		 (p[M13]*p[M21]-p[M11]*p[M23])/d,
		 (p[M11]*p[M22]-p[M12]*p[M21])/d);
}

/* inversione */
template<class T>
Mat3x3T<T> Mat3x3T<T>::Inv(void) const
{
   T d = dDet();
   if (fabs(d) < DBL_EPSILON) {
      silent_cerr("matrix is singular" << std::endl);
      throw MatrixHandler::ErrMatrixIsSingular();
   }
   
   return Inv(d);
}


/* soluzione */
template<class T>
Vec3T<T> Mat3x3T<T>::Inv(const T& d, const Vec3T<T>& v) const
{
   T* p = (T*)pdMat;
   T* pv = v.pGetVec();

   ASSERT(fabs(d) > DBL_EPSILON);
   
   return Vec3T<T>((pv[V1]*(p[M22]*p[M33]-p[M23]*p[M32])
		+pv[V2]*(p[M13]*p[M32]-p[M12]*p[M33])
		+pv[V3]*(p[M12]*p[M23]-p[M13]*p[M22]))/d,
	       (pv[V1]*(p[M23]*p[M31]-p[M21]*p[M33])
		+pv[V2]*(p[M11]*p[M33]-p[M13]*p[M31])
		+pv[V3]*(p[M13]*p[M21]-p[M11]*p[M23]))/d,
	       (pv[V1]*(p[M21]*p[M32]-p[M22]*p[M31])
		+pv[V2]*(p[M12]*p[M31]-p[M11]*p[M32])
		+pv[V3]*(p[M11]*p[M22]-p[M12]*p[M21]))/d);
}

/* soluzione */
template<class T>
Vec3T<T> Mat3x3T<T>::Inv(const Vec3T<T>& v) const
{
   T d = dDet();
   
   if (fabs(d) < DBL_EPSILON) {
      silent_cerr("matrix is singular" << std::endl);
      throw ErrGeneric();
   }

   return Inv(d, v);
}

/* Mat3x3T - end */

/* Manipolatori */
_MatR_Manip<doublereal> MatR;
_MatG_Manip<doublereal> MatG;
_MatGm1_Manip<doublereal> MatGm1;


template<class T>
Vec3T<T> operator - (const Vec3T<T>& v)
{
   return Vec3T<T>(-v.pdVec[V1],
	       -v.pdVec[V2],
	       -v.pdVec[V3]);
}


template<class T>
Mat3x3T<T> operator - (const Mat3x3T<T>& m)
{
   T* pdMat = m.pGetMat();
   return Mat3x3T<T>(-pdMat[M11],
		 -pdMat[M21],
		 -pdMat[M31],
		 -pdMat[M12],
		 -pdMat[M22],
		 -pdMat[M32],
		 -pdMat[M13],
		 -pdMat[M23],
		 -pdMat[M33]);
}


const char sForm[] = "%15.6e%15.6e%15.6e%15.6e%15.6e%15.6e%15.6e%15.6e%15.6e";
const char sDefFill[] = " ";

/* output di matrici */

template<class T>
std::ostream& 
operator << (std::ostream& out, const Mat3x3T<T>& m)
{
   T* pd = m.pGetMat();

   out
     << pd[M11] << sDefFill << pd[M12] << sDefFill << pd[M13] << sDefFill
     << pd[M21] << sDefFill << pd[M22] << sDefFill << pd[M23] << sDefFill
     << pd[M31] << sDefFill << pd[M32] << sDefFill << pd[M33];
   
   return out;
}


template<class T>
std::ostream& 
Write(std::ostream& out, const Mat3x3T<T>& m, const char* s, const char* s2)
{
   return m.Write(out, s, s2);
}


/* output di vettori */

template<class T>
std::ostream&
operator << (std::ostream& out, const Vec3T<T>& v)
{
   T* pd = v.pGetVec();
   
   out << pd[0] << sDefFill << pd[1] << sDefFill << pd[2];
   
   return out;
}


template<class T>
std::ostream& 
Write(std::ostream& out, const Vec3T<T>& v, const char* s)
{
   return v.Write(out, s);
}


/* Output di matrici */
template<class T>
std::ostream&
Mat3x3T<T>::Write(std::ostream& out, const char* sFill, const char* sFill2) const
{
   char* sF2 = (char*)sFill2; 
   if (sFill2 == NULL) {
      sF2 = (char*)sFill;
   }
   out 
     << pdMat[M11] << sFill << pdMat[M12] << sFill << pdMat[M13] << sF2
     << pdMat[M21] << sFill << pdMat[M22] << sFill << pdMat[M23] << sF2
     << pdMat[M31] << sFill << pdMat[M32] << sFill << pdMat[M33];
   
   return out;
}


template<class T>
std::ostream&
Vec3T<T>::Write(std::ostream& out, const char* sFill) const
{
   out << pdVec[V1] << sFill << pdVec[V2] << sFill << pdVec[V3];
   
   return out;
}


template<class T>
std::ostream& Write(std::ostream& out, const T& d, const char*)
{
   return out << d;
}


/* calcolo dei parametri di rotazione a partire dalla matrice R */
template<class T>
Vec3T<T> gparam(const Mat3x3T<T>& m)
{
   /* test di singolarita' */
   T d = 1.+m.Trace();
   
   if (fabs(d) < DBL_EPSILON) {
      silent_cerr("gparam(): divide by zero,"
        " probably due to singularity in rotation parameters" << std::endl);
      throw ErrGeneric();
   }
   
   return m.Ax()*(4./d);
}


/* Calcolo della matrice R a partire da due vettori sghembi */

template<class T>
Mat3x3T<T> MatR2vec(unsigned short int ia, const Vec3T<T>& va, 
		unsigned short int ib, const Vec3T<T>& vb)
{
   const char sFuncName[] = "MatR2vec()";
   
   ASSERT(ia >= 1 && ia <= 3); 
   ASSERT(va.Dot() > DBL_EPSILON);
   ASSERT(ib >= 1 && ib <= 3); 
   ASSERT(vb.Dot() > DBL_EPSILON);
   
   Vec3T<T> r[3];

   DEBUGCOUT(sFuncName << ": ia = " << ia << " (" << va << "),"
	     << " ib = " << ib << " (" << vb << ")" << std::endl);
   
   if (ia < 1 || ia > 3) {
      silent_cerr(sFuncName << ": first index is illegal" 
	      << std::endl);
      throw ErrGeneric();
   }
   
   int i1 = ia-1;
   int i2 = ia%3;
   int i3 = (ia+1)%3;
   
   if (ib == (ia%3)+1) {
      r[i1] = va/va.Norm();
      r[i3] = r[i1].Cross(vb);
      T d = r[i3].Dot();
      if (d <= DBL_EPSILON) {
	 silent_cerr(sFuncName << ": vectors must be distinct" 
		 << std::endl);
	 throw ErrGeneric();
      }	
      d = sqrt(d);
      r[i3] /= d;
      r[i2] = r[i3].Cross(r[i1]);
      
      DEBUGCOUT("R = " << Mat3x3T<T>(r[0], r[1], r[2]) << std::endl);
      
      return Mat3x3T<T>(r[0], r[1], r[2]);
   } else if (ib == ((ia+1)%3+1)) {
      r[i1] = va/va.Norm();	
      r[i2] = vb.Cross(r[i1]);
      T d = r[i2].Dot();
      if (d <= DBL_EPSILON) {
	 silent_cerr(sFuncName << ": vectors must be distinct" 
		 << std::endl);
	 throw ErrGeneric();
      }	
      d = sqrt(d);
      r[i2] /= d;
      r[i3] = r[i1].Cross(r[i2]);  
      
      DEBUGCOUT("R = " << Mat3x3<T>(r[0], r[1], r[2]) << std::endl);
      
      return Mat3x3T<T>(r[0], r[1], r[2]);
   } else {
      silent_cerr(sFuncName << ": second index is illegal" << std::endl);
      throw ErrGeneric();
   }
   
   return Mat3x3T<T>(0.); // phony call, not reachable
}


/* Subroutine che calcola gli angoli di Eulero a partire dalla matrice R
      SUBROUTINE RPYPAR(ALFA, PHI)
      
      IMPLICIT NONE
      REAL*8 PHI(3), ALFA(3,3)

      REAL*8 RADEGR
      PARAMETER (RADEGR = 180.D0/3.141592653589793D0)

      REAL*8 CA, SA

      PHI(1) = ATAN2(- ALFA(2,3), ALFA(3,3))
      CA = COS(PHI(1))
      SA = SIN(PHI(1))
      PHI(1) = PHI(1)*RADEGR
      PHI(2) = ATAN2(ALFA(1,3), - SA*ALFA(2,3) + CA*ALFA(3,3))*RADEGR
      PHI(3) = ATAN2(CA*ALFA(2,1) + SA*ALFA(3,1),
     &               CA*ALFA(2,2) + SA*ALFA(3,2))*RADEGR
      RETURN
      END

 */

const doublereal dRaDegr = 180./M_PI;

template<class T>
Vec3T<T> MatR2EulerAngles(const Mat3x3T<T>& R)
{  
   T dAlpha = atan2(-R.dGet(2,3), R.dGet(3,3));
   T dCosAlpha = cos(dAlpha);
   T dSinAlpha = sin(dAlpha);
      
   return Vec3T<T>(dAlpha,
	       atan2(R.dGet(1,3), 
		     dCosAlpha*R.dGet(3,3)-dSinAlpha*R.dGet(2,3)),
	       atan2(dCosAlpha*R.dGet(2,1)+dSinAlpha*R.dGet(3,1),
		     dCosAlpha*R.dGet(2,2)+dSinAlpha*R.dGet(3,2)));   
}

template<class T>
void MatR2EulerParams(const Mat3x3T<T>& R, T& e0, Vec3T<T>& e)
{
   T t = R.Tr();
   T Tx[4];
   Tx[0] = 1. + t;
   Tx[1] = 1. - t + 2.*R.dGet(1, 1);
   Tx[2] = 1. - t + 2.*R.dGet(2, 2);
   Tx[3] = 1. - t + 2.*R.dGet(3, 3);
   
   int k = 0;
   for (int i = 1; i <= 3; i++) {
      if (fabs(Tx[i]) >= fabs(Tx[k])) {
	 k = i;       
      }
   }
   
   switch (k) {
    case 0:   
      e0 = .5*sqrt(Tx[0]);
      e = Vec3T<T>(R.dGet(3, 2)-R.dGet(2, 3),
	       R.dGet(1, 3)-R.dGet(3, 1),
	       R.dGet(2, 1)-R.dGet(1, 2))/(4.*e0);
      break;
      
    case 1: {
       T e1 = .5*sqrt(Tx[1]);       
       e0 = (R.dGet(3, 2) - R.dGet(2, 3))/(4.*e1);
       e = Vec3T<T>(Tx[1],
		R.dGet(2, 1)+R.dGet(1, 2),
		R.dGet(3, 1)+R.dGet(1, 3))/(4.*e1);
       break;
    }
      
    case 2: {
       T e2 = .5*sqrt(Tx[2]);
       e0 = (R.dGet(1, 3) - R.dGet(3, 1))/(4.*e2);
       e = Vec3T<T>(R.dGet(1, 2)+R.dGet(2, 1),
		Tx[2],
		R.dGet(3, 2)+R.dGet(2, 3))/(4.*e2);
       break;
    }
      
    case 3: {
       T e3 = .5*sqrt(Tx[3]);
       e0 = (R.dGet(2, 1) - R.dGet(1, 2))/(4.*e3);
       e = Vec3T<T>(R.dGet(1, 3)+R.dGet(3, 1),
		R.dGet(2, 3)+R.dGet(3, 2),
		Tx[3])/(4.*e3);
       break;
    }
   }
}

template<class T>
Mat3x3T<T> EulerAngles2MatR(const Vec3T<T>& v)
{
   T d = v.dGet(1);
   T dCosAlpha(cos(d));
   T dSinAlpha(sin(d));
   d = v.dGet(2);
   T dCosBeta(cos(d));
   T dSinBeta(sin(d));
   d = v.dGet(3);
   T dCosGamma(cos(d));
   T dSinGamma(sin(d));
   
   return Mat3x3T<T>(dCosBeta*dCosGamma,
		 dCosAlpha*dSinGamma+dSinAlpha*dSinBeta*dCosGamma,
		 dSinAlpha*dSinGamma-dCosAlpha*dSinBeta*dCosGamma,
		 -dCosBeta*dSinGamma,
		 dCosAlpha*dCosGamma-dSinAlpha*dSinBeta*dSinGamma,
		 dSinAlpha*dCosGamma+dCosAlpha*dSinBeta*dSinGamma,
		 dSinBeta,
		 -dSinAlpha*dCosBeta,
		 dCosAlpha*dCosBeta);
};

template<class T>
Vec3T<T> Unwrap(const Vec3T<T>& vPrev, const Vec3T<T>& v)
{
	T dTheta = v.Norm();
	if (dTheta > 0.) {
		T dThetaPrev = vPrev.Norm();
		if (dThetaPrev > DBL_EPSILON) {
			bool b(false);

			if (vPrev*v < 0) {
				dTheta = -dTheta;
			}

			T dThetaOld = dTheta;

			while (dTheta - dThetaPrev > M_PI) {
				dTheta -= 2.*M_PI;
				b = true;
			}

			while (dThetaPrev - dTheta > M_PI) {
				dTheta += 2.*M_PI;
				b = true;
			}

			if (b) {
				return v*(dTheta/dThetaOld);
			}
		}
	}

	return v;
}


template class Vec3T<doublereal>;
template class Mat3x3T<doublereal>;
template Vec3T<doublereal> gparam<doublereal>(const Mat3x3T<doublereal>& m);
template Mat3x3T<doublereal> MatR2vec(unsigned short int ia, const Vec3T<doublereal>& va, 
			unsigned short int ib, const Vec3T<doublereal>& vb);
template Vec3T<doublereal> MatR2EulerAngles(const Mat3x3T<doublereal>& R);
template Vec3T<doublereal> Unwrap(const Vec3T<doublereal>& vPrev, const Vec3T<doublereal>& v);
template void MatR2EulerParams(const Mat3x3T<doublereal>& R, doublereal& e0,
			Vec3T<doublereal>& e);
template Mat3x3T<doublereal> EulerAngles2MatR(const Vec3T<doublereal>& v);
template Vec3T<doublereal> operator - (const Vec3T<doublereal>& v);
template Mat3x3T<doublereal> operator - (const Mat3x3T<doublereal>& m);
template std::ostream& operator << (std::ostream& out, const
			Mat3x3T<doublereal>& m);
template std::ostream& Write(std::ostream& out, const Mat3x3T<doublereal>& m, 
			const char* s, const char* s2);
template std::ostream& operator << (std::ostream& out,
			const Vec3T<doublereal>& v);
template std::ostream& Write(std::ostream& out, const Vec3T<doublereal>& v,
			const char* s);
template std::ostream& Write(std::ostream& out, const doublereal& d, const char*);


#include "CppAD/CppAD.h"

namespace CppAD {
AD<doublereal> fabs(const AD<doublereal>& f) {return abs(f);};
}

template class Vec3T<CppAD::AD<doublereal> >;
template class Mat3x3T<CppAD::AD<doublereal> >;
template Vec3T<CppAD::AD<doublereal> > gparam<CppAD::AD<doublereal> >(const Mat3x3T<CppAD::AD<doublereal> >& m);
template Mat3x3T<CppAD::AD<doublereal> > MatR2vec(unsigned short int ia, const Vec3T<CppAD::AD<doublereal> >& va, 
			unsigned short int ib, const Vec3T<CppAD::AD<doublereal> >& vb);
template Vec3T<CppAD::AD<doublereal> > MatR2EulerAngles(const Mat3x3T<CppAD::AD<doublereal> >& R);
template Vec3T<CppAD::AD<doublereal> > Unwrap(const Vec3T<CppAD::AD<doublereal> >& vPrev, const Vec3T<CppAD::AD<doublereal> >& v);
template void MatR2EulerParams(const Mat3x3T<CppAD::AD<doublereal> >& R, CppAD::AD<doublereal>& e0,
			Vec3T<CppAD::AD<doublereal> >& e);
template Mat3x3T<CppAD::AD<doublereal> > EulerAngles2MatR(const Vec3T<CppAD::AD<doublereal> >& v);
template Vec3T<CppAD::AD<doublereal> > operator - (const Vec3T<CppAD::AD<doublereal> >& v);
template Mat3x3T<CppAD::AD<doublereal> > operator - (const Mat3x3T<CppAD::AD<doublereal> >& m);
template std::ostream& operator << (std::ostream& out, const
			Mat3x3T<CppAD::AD<doublereal> >& m);
template std::ostream& Write(std::ostream& out, const Mat3x3T<CppAD::AD<doublereal> >& m, 
			const char* s, const char* s2);
template std::ostream& operator << (std::ostream& out,
			const Vec3T<CppAD::AD<doublereal> >& v);
template std::ostream& Write(std::ostream& out, const Vec3T<CppAD::AD<doublereal> >& v,
			const char* s);
template std::ostream& Write(std::ostream& out, const CppAD::AD<doublereal>& d, const char*);

