/* 
 * MBDyn (C) is a multibody analysis code. 
 * http://www.mbdyn.org
 *
 * Copyright (C) 1996-2005
 *
 * Pierangelo Masarati	<masarati@aero.polimi.it>
 * Paolo Mantegazza	<mantegazza@aero.polimi.it>
 *
 * Dipartimento di Ingegneria Aerospaziale - Politecnico di Milano
 * via La Masa, 34 - 20156 Milano, Italy
 * http://www.aero.polimi.it
 *
 * Changing this copyright notice is forbidden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation (version 2 of the License).
 * 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* vettori e matrici 3x3 - operazioni collegate */


#ifndef MATVEC3T_H
#define MATVEC3T_H

#include <cmath>
#include <ac/iostream>
#include <ac/f2c.h>

#include <myassert.h>
#include <except.h>
#include <solman.h>

enum {
   V1 = 0,
   V2 = 1,
   V3 = 2
};

enum {
   M11 = 0,
   M12 = 3,
   M13 = 6,
   M21 = 1,
   M22 = 4,
   M23 = 7,
   M31 = 2,
   M32 = 5,
   M33 = 8
};

/* classi principali definite in questo file */
template<class T> class Vec3T;      /* vettore 3x1 */
template<class T> class Mat3x3T;    /* matrice 3x3 */

template<class T> class _Mat3x3T_Manip;  /* manipolatori per la matrice 3x3 */
template<class T> Vec3T<T> operator - (const Vec3T<T>& v);


/* Vec3T - begin */

// Vettori di dimensione 3
template<class T>
class Vec3T {
   friend class Mat3x3T<T>;   
   friend Vec3T<T> operator -<> (const Vec3T<T>& v);
   friend class _Mat3x3T_Manip<T>;
   friend class VecN;
   friend class Mat3xN;
   friend class MatNx3;
   

 private:
   //vettore di tre reali che contiene i coefficienti
   T pdVec[3];

 protected:
   
 public:
   /*Costruttori */

   /* 
    Costruttore banalissimo: non fa nulla. Attenzione che cosi' il
    valore di Vec3T e' unpredictable. Se si desidera un vettore nullo
    usare Vec3T(0.);
    */   
   Vec3T(void) { 
      NO_OP; 
   };
   
   /*
    Assegna i tre valori. Se gli ultimi due non sono presenti li setta a zero.
    Per azzerare il vettore, basta usare Vec3T(0.)
    */
   Vec3T(const T& v1, 
	const T& v2 = 0., 
	const T& v3 = 0.) {
      pdVec[V1] = v1; 
      pdVec[V2] = v2; 
      pdVec[V3] = v3; 
   };
   
   /*
    Costruttore di copia.
    */
   template<class TT>
   Vec3T(const Vec3T<TT>& v) {
      pdVec[V1] = v[V1];
      pdVec[V2] = v[V2];
      pdVec[V3] = v[V3];
   };
   
   /*
    Costruttore da array di reali. Copia i primi 3 valori dell'array.
    Si assume che l'array da pd sia lungo almeno 3. 
    */
   Vec3T(const T* pd) {
      ASSERT(pd != NULL);
      
      pdVec[V1] = pd[0];
      pdVec[V2] = pd[1];
      pdVec[V3] = pd[2];
   };
   
   /*
    Costruttore da VectorHandler. Prende i valori da iFirstIndex 
    a iFirstIndex+1. Nota: gli indici del VectorHandler partono da 1, 
    in stile FORTRAN.
    */
   Vec3T(const VectorHandler& vh, integer iFirstIndex) {
      ASSERT(iFirstIndex > 0 && iFirstIndex <= vh.iGetSize()-2);
      pdVec[V1] = vh.dGetCoef(iFirstIndex);
      pdVec[V2] = vh.dGetCoef(++iFirstIndex);
      pdVec[V3] = vh.dGetCoef(++iFirstIndex);
   };
   
   /*
    Distruttore banale: non fa nulla.
    */
   ~Vec3T(void) { 
      NO_OP; 
   };
      

   /*Metodi di servizio */
      
   /*
    Dirty job: restituisce il puntatore al vettore (deprecato).
    */
   T* pGetVec(void) const { 
      return (T*)pdVec; 
   };
      
   /*Operatori su vettori e matrici */
      
   /*
    Prodotto "tensore". 
    Restituisce se stesso per v trasposto.
    */
   Mat3x3T<T> Tens(const Vec3T& v) const;
      
   /*
    Prodotto "tensore". 
    Restituisce se stesso per se stesso
    */
   Mat3x3T<T> Tens(void) const;
 
   /*
    Prodotto vettore. 
    Restituisce il prodotto vettore tra se stesso e v in un temporaneo.
    */
   Vec3T Cross(const Vec3T& v) const {     
      return Vec3T(pdVec[V2]*v.pdVec[V3]-pdVec[V3]*v.pdVec[V2],
		  pdVec[V3]*v.pdVec[V1]-pdVec[V1]*v.pdVec[V3],
		  pdVec[V1]*v.pdVec[V2]-pdVec[V2]*v.pdVec[V1]);
   };
   
   /*
    Prodotto vettore per matrice. 
    Restituisce il prodotto vettore tra se stesso e m in un temporaneo.
    */
   Mat3x3T<T> Cross(const Mat3x3T<T>& m) const;
   
   /*
    Prodotto scalare. 
    Restituisce il prodotto scalare tra se e v
    */
   T Dot(const Vec3T& v) const {
      return 
	pdVec[V1]*v.pdVec[V1]+
	pdVec[V2]*v.pdVec[V2]+
	pdVec[V3]*v.pdVec[V3];
   };
   
   /*
    Prodotto scalare per se stesso.
    */
   T Dot(void) const { 
      return 
	pdVec[V1]*pdVec[V1]+
	pdVec[V2]*pdVec[V2]+
	pdVec[V3]*pdVec[V3];
   };
   
   /*
    Norma: sqrt(Dot())
    */
   T Norm(void) const { 
      return 
	sqrt(pdVec[V1]*pdVec[V1]+
	     pdVec[V2]*pdVec[V2]+
	     pdVec[V3]*pdVec[V3]);
   };

   /*Operazioni sui coefficienti */
      
   /*
    Assegnazione di un coefficiente. 
    Nota: l'indice ha base 1, in stile FORTRAN.
    */
   inline void Put(unsigned short int iRow, const T& dCoef) {
      ASSERT(iRow >= 1 && iRow <= 3);   
      pdVec[--iRow] = dCoef;
   };
   
   /*
    Lettura di un coefficiente.
    Nota: l'indice ha base 1, in stile FORTRAN.
    */
   inline const T& dGet(unsigned short int iRow) const {
      ASSERT(iRow >= 1 && iRow <= 3);
      return pdVec[--iRow];
   };

   inline T& operator () (unsigned short int iRow) {
      ASSERT(iRow >= 1 && iRow <= 3);
      return pdVec[--iRow];
   };

   inline const T& operator () (unsigned short int iRow) const {
      ASSERT(iRow >= 1 && iRow <= 3);
      return pdVec[--iRow];
   };

   inline T& operator [] (unsigned short int iRow) {
      ASSERT(iRow <= 2);
      return pdVec[iRow];
   };

   inline const T& operator [] (unsigned short int iRow) const {
      ASSERT(iRow <= 2);
      return pdVec[iRow];
   };
   
   
   /*Operazioni con array di reali */
      
   /*
    Somma se stesso all'array pd.
    Si assume che l'array pd sia lungo almeno 3 
    */
   void AddTo(T* pd) const {	
      ASSERT(pd != NULL);
      pd[0] += pdVec[V1];
      pd[1] += pdVec[V2];
      pd[2] += pdVec[V3];
   };
   
   /*
    Sottrae se stesso dall'array pd.
    Si assume che l'array pd sia lungo almeno 3 
    */
   void SubFrom(T* pd) const {	
      ASSERT(pd != NULL);
      pd[0] -= pdVec[V1];
      pd[1] -= pdVec[V2];
      pd[2] -= pdVec[V3];
   };
   
   /*
    Scrive se stesso sull'array pd.
    Si assume che l'array pd sia lungo almeno 3 
    */
   void PutTo(T* pd) const {
      ASSERT(pd != NULL);
      pd[0] = pdVec[V1];
      pd[1] = pdVec[V2];
      pd[2] = pdVec[V3];
   };   
   
   /*
    Si legge dall'array pd.
    Si assume che l'array pd sia lungo almeno 3
    */
   void GetFrom(const T* pd) {
      ASSERT(pd != NULL);
      pdVec[V1] = pd[0];
      pdVec[V2] = pd[1];
      pdVec[V3] = pd[2];
   };

   /*Operatori */
      
   /*
    Operatore di assegnazione 
    */
   const Vec3T& operator = (const Vec3T& v) {
      pdVec[V1] = v.pdVec[V1];
      pdVec[V2] = v.pdVec[V2];
      pdVec[V3] = v.pdVec[V3];
      
      return *this;
   };
      
   /*
    Operatore somma. 
    Restituisce v sommato a se stesso in un temporaneo.
    */
   Vec3T operator + (const Vec3T& v) const {
      return Vec3T(pdVec[V1]+v.pdVec[V1],
		  pdVec[V2]+v.pdVec[V2],
		  pdVec[V3]+v.pdVec[V3]);
   };
   
   /*
    Operatore somma e assegnazione. 
    Somma v a se stesso in loco.
    */
   const Vec3T& operator += (const Vec3T& v) {
      pdVec[V1] += v.pdVec[V1];
      pdVec[V2] += v.pdVec[V2];
      pdVec[V3] += v.pdVec[V3];
      return *this;
   };
   
   /*
    Operatore sottrazione. 
    Sottrae v da se stesso in un temporaneo.
    */
   Vec3T operator - (const Vec3T& v) const {
      return Vec3T(pdVec[V1]-v.pdVec[V1],
		  pdVec[V2]-v.pdVec[V2],
		  pdVec[V3]-v.pdVec[V3]);
   };
   
   /*
    Operatore sottrazione e assegnazione. 
    Sottrae v da se stesso in loco.
    */
   const Vec3T& operator -= (const Vec3T& v) {
      pdVec[V1] -= v.pdVec[V1];
      pdVec[V2] -= v.pdVec[V2];
      pdVec[V3] -= v.pdVec[V3];
      return *this;
   };
   
   /*
    Operatore prodotto per scalare. 
    Moltiplica se stesso per d in un temporaneo.
    */
   Vec3T operator * (const T& d) const {
      return Vec3T(pdVec[V1]*d,
		  pdVec[V2]*d,
		  pdVec[V3]*d);    
   };
   
   /*
    Operatore prodotto e assegnazione per scalare.
    Moltiplica se stesso per d in loco.
    */
   const Vec3T& operator *= (const T& d) {
      pdVec[V1] *= d;
      pdVec[V2] *= d;
      pdVec[V3] *= d;
      return *this;
   };   

   /*
    Prodotto scalare. 
    Moltiplica se stesso per v.
    */
   T operator * (const Vec3T& v) const {
      return pdVec[V1]*v.pdVec[V1]
	+pdVec[V2]*v.pdVec[V2]
	+pdVec[V3]*v.pdVec[V3];
   };   
   
   /*
    Operatore divisione per scalare. 
    Divide se stesso per d in un temporaneo.
    */
   Vec3T operator / (const T& d) const {
      ASSERT(d != 0.);      
      return Vec3T(pdVec[V1]/d,
		  pdVec[V2]/d,
		  pdVec[V3]/d);    
   };
   
   /*
    Operatore divisione e assegnazione per scalare. 
    Divide se stesso per d in loco.
    */
   const Vec3T& operator /= (const T& d) {
      ASSERT(d != 0.);	
      pdVec[V1] /= d;
      pdVec[V2] /= d;
      pdVec[V3] /= d;
      return *this;
   };   
   
   /*
    Operatore booleano di uguaglianza tra vettori.
    */
      bool operator == (const Vec3T& v) const {
	 return (pdVec[V1] == v.pdVec[V1] 
		 && pdVec[V2] == v.pdVec[V2] 
		 && pdVec[V3] == v.pdVec[V3]);
      };

   /*
    Operatore booleano di disuguaglianza tra vettori.
    */
      bool operator != (const Vec3T& v) const {
	 return (pdVec[V1] != v.pdVec[V1] 
		 || pdVec[V2] != v.pdVec[V2] 
		 || pdVec[V3] != v.pdVec[V3]);
      };
     
   /*Input/Output */
      
   /*
    Scrive se stesso sull'ostream out.
    I coefficienti sono separati dalla stringa sFill (spazio di default).
    */
   std::ostream& Write(std::ostream& out, const char* sFill = " ") const;
};
   
/* Vec3T - end */


/* _Mat3x3T_Manip - begin */

/* classe virtuale dei manipolatori */
/* 
 Manipolatori per la costruzione di matrici 3x3.
 Sono usati per ottenere le matrici di rotazione e lel loro derivate
 dai parametri di rotazione
 */
template<class T>
class _Mat3x3T_Manip {
 public:
   virtual ~_Mat3x3T_Manip() {NO_OP;};
   /*
    Operatore deprecato (e' poco efficiente).
    L'uso e': m = _Mat3x3T_Manip << v;
    dove m e' una Mat3x3T, v e' un Vec3T e _Mat3x3T_Manip e' un oggetto
    derivato da questa classe che implementa la trasformazione da
    vettore a matrice.
    */
   virtual Mat3x3T<T> operator << (const Vec3T<T>& v) const = 0;
   
   /*
    Metodo che trasforma il vettore v nella matrice m.
    Viene usato da un costruttore di Mat3x3T che riceve come
    argomenti un manipolatore e un vettore di parametri di rotazione.
    */
   virtual void Make(Mat3x3T<T>& m, const Vec3T<T>& v) const = 0;
};

/* _Mat3x3T_Manip - end */


/* Mat3x3T - begin */
// Matrici 3x3
template<class T>
class Mat3x3T {
   friend class Vec3T<T>;
   friend class SparseSubMatrixHandler;
   friend class _Mat3x3T_Manip<T>;   
   friend class Mat3xN;
   friend class MatNx3;
   
 protected:
   //Vettore di 9 reali che contiene i coefficienti
   T pdMat[9];
   
 public:

   /*Costruttori */
   /*
    Costruttore banale: non inizializza i coefficienti.
    Per azzerare la matrice usare Mat3x3T(0.)
    */
   Mat3x3T(void) {
      NO_OP;
   };
   
   /* Costruttore di matrice diagonale.
    Scrive il valore in ingresso sulla diagonale principale. 
    Di conseguenza, se e' nullo automaticamente azzera la matrice.
    */
   Mat3x3T(const T& d) {
      pdMat[M11] = d;
      pdMat[M21] = 0.;
      pdMat[M31] = 0.;
      pdMat[M12] = 0;
      pdMat[M22] = d;
      pdMat[M32] = 0.;
      pdMat[M13] = 0.;
      pdMat[M23] = 0.;
      pdMat[M33] = d;
   };
         
   /*
    Costrutture completo.
    */
   Mat3x3T(const T& m11, const T& m21, const T& m31,
	  const T& m12, const T& m22, const T& m32,
	  const T& m13, const T& m23, const T& m33) {
      pdMat[M11] = m11;
      pdMat[M21] = m21;
      pdMat[M31] = m31;
      pdMat[M12] = m12;
      pdMat[M22] = m22;
      pdMat[M32] = m32;
      pdMat[M13] = m13;
      pdMat[M23] = m23;
      pdMat[M33] = m33;      
   };
   
   /*
    Costruttore di copia.
    */
   template<class TT>
   Mat3x3T(const Mat3x3T<TT>& m) {
      pdMat[M11] = m.pGetMat()[M11];
      pdMat[M21] = m.pGetMat()[M21];
      pdMat[M31] = m.pGetMat()[M31];
      pdMat[M12] = m.pGetMat()[M12];
      pdMat[M22] = m.pGetMat()[M22];
      pdMat[M32] = m.pGetMat()[M32];
      pdMat[M13] = m.pGetMat()[M13];
      pdMat[M23] = m.pGetMat()[M23];
      pdMat[M33] = m.pGetMat()[M33];
   };
   
   /*
    Costruttore prodotto vettore.
    La matrice viene inizializzata con il vettore v disposto a dare 
    la matrice prodotto vettore 
    (ovvero la matrice che, moltiplicata per w, da' v vettor w).
    */
   Mat3x3T(const Vec3T<T>& v) {
      pdMat[M11] = 0.;
      pdMat[M21] = v.pdVec[V3];
      pdMat[M31] = -v.pdVec[V2];
      pdMat[M12] = -v.pdVec[V3];
      pdMat[M22] = 0.;      
      pdMat[M32] = v.pdVec[V1];
      pdMat[M13] = v.pdVec[V2];
      pdMat[M23] = -v.pdVec[V1];
      pdMat[M33] = 0.;
   };
   
   /*
    Costruttore doppio prodotto vettore.
    restituisce la matrice data dal prodotto di due matrici prodotto vettore.
    */
   Mat3x3T(const Vec3T<T>& a, const Vec3T<T>& b) {
      pdMat[M11] = -a.pdVec[V2]*b.pdVec[V2]-a.pdVec[V3]*b.pdVec[V3];
      pdMat[M21] = a.pdVec[V1]*b.pdVec[V2];
      pdMat[M31] = a.pdVec[V1]*b.pdVec[V3];
      pdMat[M12] = a.pdVec[V2]*b.pdVec[V1];
      pdMat[M22] = -a.pdVec[V3]*b.pdVec[V3]-a.pdVec[V1]*b.pdVec[V1];
      pdMat[M32] = a.pdVec[V2]*b.pdVec[V3];
      pdMat[M13] = a.pdVec[V3]*b.pdVec[V1];
      pdMat[M23] = a.pdVec[V3]*b.pdVec[V2];
      pdMat[M33] = -a.pdVec[V1]*b.pdVec[V1]-a.pdVec[V2]*b.pdVec[V2];
   };
   
   
   /*
    Costruttore che piglia tre vettori e li affianca a dare la matrice
    */
   Mat3x3T(const Vec3T<T>& v1, const Vec3T<T>& v2, const Vec3T<T>& v3) {
      pdMat[M11] = v1.pdVec[V1];
      pdMat[M21] = v1.pdVec[V2];
      pdMat[M31] = v1.pdVec[V3];

      pdMat[M12] = v2.pdVec[V1];
      pdMat[M22] = v2.pdVec[V2];
      pdMat[M32] = v2.pdVec[V3];

      pdMat[M13] = v3.pdVec[V1];
      pdMat[M23] = v3.pdVec[V2];
      pdMat[M33] = v3.pdVec[V3];
   };
   
   
   /*
    Costruttore di copia da array.
    si assume che l'array pd sia lungo almeno 9 
    */
   Mat3x3T(const T* pd, integer iSize) {
      ASSERT(pd != NULL);      
      GetFrom(pd, iSize);
   };   
   
   /*
    Costruttore con manipolatore. 
    Invoca un metodo del manipolatore Manip che restituisce una matrice 
    a partire dal vettore v.
    */
   Mat3x3T(const _Mat3x3T_Manip<T>& Manip, const Vec3T<T>& v) {
      Manip.Make(*this, v);
   };
   
   
   /*
    Costruttore che genera la matrice identita' moltiplicata per d e sommata
    alla matrice prodotto vettore ottenuta da v.
    */
   Mat3x3T(const T& d, const Vec3T<T>& v) {
      pdMat[M11] = d;
      pdMat[M21] = v.pdVec[V3];
      pdMat[M31] = -v.pdVec[V2];
      pdMat[M12] = -v.pdVec[V3];
      pdMat[M22] = d;
      pdMat[M32] = v.pdVec[V1];
      pdMat[M13] = v.pdVec[V2];
      pdMat[M23] = -v.pdVec[V1];
      pdMat[M33] = d;
   };
   
   /*
    Distruttore banale.
    */
   ~Mat3x3T(void) { 
      NO_OP;
   };


   /*Metodi di servizio */
      
   /*
    Dirty job: restituisce il puntatore alla matrice (deprecato).
    */
   T* pGetMat(void) const { 
      return (T*)pdMat;
   };

      
   /*Operazioni sui coefficienti */
      
   /*
    Assegnazione di un coefficiente.
    Nota: gli indici hanno base 1, in stile FORTRAN.
    */
   inline void Put(unsigned short int iRow,
		   unsigned short int iCol, 
		   const T& dCoef) {
      ASSERT(iRow >= 1 && iRow <= 3);
      ASSERT(iCol >= 1 && iCol <= 3);      
      pdMat[--iRow+3*--iCol] = dCoef;
   };
   
   /*
    Lettura di un coefficiente.
    Nota: gli indici hanno base 1, in stile FORTRAN.
    */
   inline const T& dGet(unsigned short int iRow, 
				 unsigned short int iCol) const {
      ASSERT(iRow >= 1 && iRow <= 3);
      ASSERT(iCol >= 1 && iCol <= 3);      
      return pdMat[--iRow+3*--iCol];
   };

   inline T& operator () (unsigned short int iRow, 
		   unsigned short int iCol) {
       ASSERT(iRow >= 1 && iRow <= 3);
       ASSERT(iCol >= 1 && iCol <= 3);
       return pdMat[--iRow+3*--iCol];
   };

   inline const T& operator () (unsigned short int iRow, 
		   unsigned short int iCol) const {
       ASSERT(iRow >= 1 && iRow <= 3);
       ASSERT(iCol >= 1 && iCol <= 3);
       return pdMat[--iRow+3*--iCol];
   };

   
   /*Operazioni su matrici e vettori */
      
   /*
    Prodotto di matrici prodotto vettore.
    Setta se stessa pari al prodotto delle matrici prodotto vettore 
    ottenute dai vettori a e b.
    */
   const Mat3x3T& Tens(const Vec3T<T>& a, const Vec3T<T>& b) {
      pdMat[M11] = a.pdVec[V1]*b.pdVec[V1];   /* m(1,1) = a(1)*b(1) */
      pdMat[M21] = a.pdVec[V2]*b.pdVec[V1];   /* m(2,1) = a(2)*b(1) */
      pdMat[M31] = a.pdVec[V3]*b.pdVec[V1];   /* m(3,1) = a(3)*b(1) */
      pdMat[M12] = a.pdVec[V1]*b.pdVec[V2];   /* m(1,2) = a(1)*b(2) */
      pdMat[M22] = a.pdVec[V2]*b.pdVec[V2];   /* m(2,2) = a(2)*b(2) */
      pdMat[M32] = a.pdVec[V3]*b.pdVec[V2];   /* m(3,2) = a(3)*b(2) */
      pdMat[M13] = a.pdVec[V1]*b.pdVec[V3];   /* m(1,3) = a(1)*b(3) */
      pdMat[M23] = a.pdVec[V2]*b.pdVec[V3];   /* m(2,3) = a(2)*b(3) */
      pdMat[M33] = a.pdVec[V3]*b.pdVec[V3];   /* m(3,3) = a(3)*b(3) */
      return *this;
   };

   /*
    Traspone la matrice.
    Restituisce la trasposta di se stessa in un temporaneo (poco efficiente).
    */
   Mat3x3T Transpose(void) const {
      /* La funzione non e' ottimizzata. Genera una nuova matrice che 
       * costruisce in modo opportuno. Se si ritiene di dover usare 
       * ripetutamente la matrice trasposta, conviene memorizzarla in
       * una variabile di servizio */
      
      return Mat3x3T(pdMat[M11], pdMat[M12], pdMat[M13],
		    pdMat[M21], pdMat[M22], pdMat[M23],
		    pdMat[M31], pdMat[M32], pdMat[M33]);
   };

   Vec3T<T> Ax(void) const {
      /* assumendo che sia una matrice di rotazione ?!? */
      return Vec3T<T>(
		      .5*(pdMat[M32]-pdMat[M23]),
		      .5*(pdMat[M13]-pdMat[M31]),
		      .5*(pdMat[M21]-pdMat[M12])
		      );
   };

   T Trace(void) const {
      return pdMat[M11]+pdMat[M22]+pdMat[M33];
   };

   Mat3x3T Symm(void) const {
      T m12 = .5*(pdMat[M21]+pdMat[M12]);
      T m13 = .5*(pdMat[M31]+pdMat[M13]);
      T m23 = .5*(pdMat[M32]+pdMat[M23]);

      return Mat3x3T(
		      pdMat[M11], m12, m13,
		      m12, pdMat[M22], m23,
		      m13, m23, pdMat[M33]
		      );
   };

   Mat3x3T Symm2(void) const {
      T m12 = pdMat[M21]+pdMat[M12];
      T m13 = pdMat[M31]+pdMat[M13];
      T m23 = pdMat[M32]+pdMat[M23];

      return Mat3x3T(
		      2.*pdMat[M11], m12, m13,
		      m12, 2.*pdMat[M22], m23,
		      m13, m23, 2.*pdMat[M33]
		      );
   };

   const Mat3x3T& Symm(const Mat3x3T& m) {
      pdMat[M11] = m.pdMat[M11];
      pdMat[M22] = m.pdMat[M22];
      pdMat[M33] = m.pdMat[M33];
      pdMat[M12] = pdMat[M21] = .5*(m.pdMat[M21]+m.pdMat[M12]);
      pdMat[M13] = pdMat[M31] = .5*(m.pdMat[M31]+m.pdMat[M13]);
      pdMat[M23] = pdMat[M32] = .5*(m.pdMat[M32]+m.pdMat[M23]);

      return *this;
   };

   Mat3x3T Skew(void) const {
      return Mat3x3T(this->Ax());
   };

   const Mat3x3T& Skew(const Mat3x3T& m) {
      pdMat[M11] = pdMat[M22] = pdMat[M13] = 0.;
      pdMat[M12] = m.pdMat[M12] - m.pdMat[M21];
      pdMat[M21] = -pdMat[M12];
      pdMat[M13] = m.pdMat[M13] - m.pdMat[M31];
      pdMat[M31] = -pdMat[M13];
      pdMat[M23] = m.pdMat[M23] - m.pdMat[M32];
      pdMat[M32] = -pdMat[M23];

      return *this;
   };

   /*
    Ottiene un sottovettore dalla matrice.
    Nota: l'indice e' a base 1, in stile FORTRAN.
    */
   Vec3T<T> GetVec(unsigned short int i) const {
      ASSERT(i >= 1 && i <= 3);
      return Vec3T<T>(pdMat+3*--i);
   };

   /*
    Ottiene un sottovettore dalla matrice.
    Nota: l'indice e' a base 1, in stile FORTRAN.
    Alias di GetVec()
    */
   Vec3T<T> GetCol(unsigned short int i) const {
      ASSERT(i >= 1 && i <= 3);
      return Vec3T<T>(pdMat + 3*--i);
   };

   /*
    Ottiene un sottovettore dalla matrice.
    Nota: l'indice e' a base 1, in stile FORTRAN.
    */
   Vec3T<T> GetRow(unsigned short int i) const {
      ASSERT(i >= 1 && i <= 3);
      --i;
      return Vec3T<T>(pdMat[i], pdMat[3 + i], pdMat[6 + i]);
   };

   void PutVec(unsigned short int i, const Vec3T<T>& v) {
      ASSERT(i >= 1 && i <= 3);

      i--; i = 3*i;
      pdMat[i++] = v.pdVec[V1];
      pdMat[i++] = v.pdVec[V2];
      pdMat[i] = v.pdVec[V3];
   };

   void AddVec(unsigned short int i, const Vec3T<T>& v) {
      ASSERT(i >= 1 && i <= 3);

      i--; i = 3*i;
      pdMat[i++] += v.pdVec[V1];
      pdMat[i++] += v.pdVec[V2];
      pdMat[i] += v.pdVec[V3];
   };

   void SubVec(unsigned short int i, const Vec3T<T>& v) {
      ASSERT(i >= 1 && i <= 3);

      i--; i = 3*i;
      pdMat[i++] -= v.pdVec[V1];
      pdMat[i++] -= v.pdVec[V2];
      pdMat[i] -= v.pdVec[V3];
   };

   /*
    Inversione. 
    Restituisce l'inversa di se stessa in un temporaneo.
    */
   T dDet(void) const;
   Mat3x3T Inv(const T& ddet) const;
   Mat3x3T Inv(void) const;
   
   /*
    Soluzione.
    Restituisce l'inversa di se stessa per v in un temporaneo.
    */
   Vec3T<T> Inv(const Vec3T<T>& v) const;      
   Vec3T<T> Inv(const T& d, const Vec3T<T>& v) const;      
   
   /*Operazioni su arrays di reali */
      
   /*
    Si legge da un array.
    Si assume che l'array pd sia lungo almeno 9.
    @param iSize e' il numero di righe di dimensionamento dell'array, 
    in stile FORTRAN
    */
   void GetFrom(const T* pd, integer iSize) {
      ASSERT(pd != NULL);
      ASSERT(iSize >= 3);

      T* pdFrom = (T*)pd;
      pdMat[M11] = pdFrom[0];
      pdMat[M21] = pdFrom[1];
      pdMat[M31] = pdFrom[2];
      
      pdFrom += iSize;
      pdMat[M12] = pdFrom[0];
      pdMat[M22] = pdFrom[1];
      pdMat[M32] = pdFrom[2];
      
      pdFrom += iSize;
      pdMat[M13] = pdFrom[0];
      pdMat[M23] = pdFrom[1];
      pdMat[M33] = pdFrom[2];
   };
   
   /*
    Somma se stesso ad un array con iNRows righe.
    si assume che l'array sia lungo almeno 2*iNRows+3 a partire da pd.
    @param iNRows e' il numero di righe di dimensionamento dell'array, 
    in stile FORTRAN
    */
   void AddTo(T* pd, integer iNRows) const {
      ASSERT(pd != NULL);
      ASSERT(iNRows >= 3);
      
      T* pdTo = pd;
      
      pdTo[0] += pdMat[M11];
      pdTo[1] += pdMat[M21];
      pdTo[2] += pdMat[M31];
      
      pdTo += iNRows;
      pdTo[0] += pdMat[M12];
      pdTo[1] += pdMat[M22];
      pdTo[2] += pdMat[M32];
      
      pdTo += iNRows;
      pdTo[0] += pdMat[M13];
      pdTo[1] += pdMat[M23];
      pdTo[2] += pdMat[M33];      
   };
   
   /*
    Copia se stesso su un array con iNRows righe.
    Si assume che l'array sia lungo almeno 2*iNRows+3.
    @param iNRows e' il numero di righe di dimensionamento dell'array, 
    in stile FORTRAN
    */
   void PutTo(T* pd, integer iNRows) const {
      ASSERT(pd != NULL);
      ASSERT(iNRows >= 3);
      
      T* pdTo = pd;
      
      pdTo[0] = pdMat[M11];
      pdTo[1] = pdMat[M21];
      pdTo[2] = pdMat[M31];
      
      pdTo += iNRows;
      pdTo[0] = pdMat[M12];
      pdTo[1] = pdMat[M22];
      pdTo[2] = pdMat[M32];
      
      pdTo += iNRows;
      pdTo[0] = pdMat[M13];
      pdTo[1] = pdMat[M23];
      pdTo[2] = pdMat[M33];
   };
   
   /*Operatori */
   
   /*
    Operatore di assegnazione.
    */
   const Mat3x3T& operator = (const Mat3x3T& m) {
      
      pdMat[M11] = m.pdMat[M11];
      pdMat[M21] = m.pdMat[M21];
      pdMat[M31] = m.pdMat[M31];
      pdMat[M12] = m.pdMat[M12];
      pdMat[M22] = m.pdMat[M22];
      pdMat[M32] = m.pdMat[M32];
      pdMat[M13] = m.pdMat[M13];
      pdMat[M23] = m.pdMat[M23];
      pdMat[M33] = m.pdMat[M33];
      
      return *this;
   };
   
   /*
    Operatore somma. 
    Restituisce v sommato a se stesso in un temporaneo.
    */
   Mat3x3T operator + (const Mat3x3T& m) const {
      return Mat3x3T(pdMat[M11]+m.pdMat[M11],
		    pdMat[M21]+m.pdMat[M21],
		    pdMat[M31]+m.pdMat[M31],
		    pdMat[M12]+m.pdMat[M12],
		    pdMat[M22]+m.pdMat[M22],
		    pdMat[M32]+m.pdMat[M32],
		    pdMat[M13]+m.pdMat[M13],
		    pdMat[M23]+m.pdMat[M23],
		    pdMat[M33]+m.pdMat[M33]);
   };
   
   /*
    Operatore somma e assegnazione.
    Somma v a se stesso in loco.
    */
   const Mat3x3T& operator += (const Mat3x3T& m) {
      pdMat[M11] += m.pdMat[M11];
      pdMat[M21] += m.pdMat[M21];
      pdMat[M31] += m.pdMat[M31];
      pdMat[M12] += m.pdMat[M12];
      pdMat[M22] += m.pdMat[M22];
      pdMat[M32] += m.pdMat[M32];
      pdMat[M13] += m.pdMat[M13];
      pdMat[M23] += m.pdMat[M23];
      pdMat[M33] += m.pdMat[M33];
      
      return *this;
   };
   
   /*
    Operatore differenza.
    Restituisce v sottratto da se stesso in un temporaneo.
    */
   Mat3x3T operator - (const Mat3x3T& m) const {
      return Mat3x3T(pdMat[M11]-m.pdMat[M11],
		    pdMat[M21]-m.pdMat[M21],
		    pdMat[M31]-m.pdMat[M31],
		    pdMat[M12]-m.pdMat[M12],
		    pdMat[M22]-m.pdMat[M22],
		    pdMat[M32]-m.pdMat[M32],
		    pdMat[M13]-m.pdMat[M13],
		    pdMat[M23]-m.pdMat[M23],
		    pdMat[M33]-m.pdMat[M33]);
   };
   
   /*
    Operatore differenza e assegnazione.
    Sottrae v da se stesso in loco.
    */
   const Mat3x3T& operator -= (const Mat3x3T& m) {
      pdMat[M11] -= m.pdMat[M11];
      pdMat[M21] -= m.pdMat[M21];
      pdMat[M31] -= m.pdMat[M31];
      pdMat[M12] -= m.pdMat[M12];
      pdMat[M22] -= m.pdMat[M22];
      pdMat[M32] -= m.pdMat[M32];
      pdMat[M13] -= m.pdMat[M13];
      pdMat[M23] -= m.pdMat[M23];
      pdMat[M33] -= m.pdMat[M33];

      return *this;
   };
   
   /*
    Operatore moltiplicazione per scalare.
    Restituisce se stesso moltiplicato per d in un temporaneo.
    */
   Mat3x3T operator * (const T& d) const {
      if (d != 1.) {
	 return Mat3x3T(pdMat[M11]*d,
		       pdMat[M21]*d,
		       pdMat[M31]*d,
		       pdMat[M12]*d,
		       pdMat[M22]*d,
		       pdMat[M32]*d,
		       pdMat[M13]*d,
		       pdMat[M23]*d,
		       pdMat[M33]*d);
      }
      return *this;
   };
   
   /*
    Operatore moltiplicazione per scalare e assegnazione.
    Moltiplica se stesso per d in loco.
    */
   const Mat3x3T& operator *= (const T& d) {
      if (d != 1.) {
	 pdMat[M11] *= d;
	 pdMat[M21] *= d;
	 pdMat[M31] *= d;
	 pdMat[M12] *= d;
	 pdMat[M22] *= d;
	 pdMat[M32] *= d;
	 pdMat[M13] *= d;
	 pdMat[M23] *= d;
	 pdMat[M33] *= d;
      }
      
      return *this;
   };
   
   /*
    Operatore divisione per scalare.
    Restituisce se stesso diviso per d in un temporaneo.
    */
   Mat3x3T operator / (const T& d) const {
      ASSERT(d != 0.);
      if (d != 1.) {
	 return Mat3x3T(pdMat[M11]/d,
		       pdMat[M21]/d,
		       pdMat[M31]/d,
		       pdMat[M12]/d,
		       pdMat[M22]/d,
		       pdMat[M32]/d,
		       pdMat[M13]/d,
		       pdMat[M23]/d,
		       pdMat[M33]/d);
      }
      
      return *this;
   };
   
   /*
    Operatore divisione per scalare e assegnazione.
    Divide se stesso per d in loco
    */
   Mat3x3T& operator /= (const T& d) {
      ASSERT(d != 0.);
      if (d != 1.) {
	 pdMat[M11] /= d;
	 pdMat[M21] /= d;
	 pdMat[M31] /= d;
	 pdMat[M12] /= d;
	 pdMat[M22] /= d;
	 pdMat[M32] /= d;
	 pdMat[M13] /= d;
	 pdMat[M23] /= d;
	 pdMat[M33] /= d;
      }
      
      return *this;
   };
   
   
   /*
    Operatore prodotto matrice vettore.
    Restituisce se stesso moltiplicato per v in un temporaneo.
    */
   Vec3T<T> operator * (const Vec3T<T>& v) const {

      return Vec3T<T>(pdMat[M11]*v.pdVec[V1]+pdMat[M12]*v.pdVec[V2]+pdMat[M13]*v.pdVec[V3],
		  pdMat[M21]*v.pdVec[V1]+pdMat[M22]*v.pdVec[V2]+pdMat[M23]*v.pdVec[V3],
		  pdMat[M31]*v.pdVec[V1]+pdMat[M32]*v.pdVec[V2]+pdMat[M33]*v.pdVec[V3]);
   };

   /*
    */
    bool operator == (const Mat3x3T& m) const {
      return pdMat[M11] == m.pdMat[M11]
	      && pdMat[M12] == m.pdMat[M12]
	      && pdMat[M13] == m.pdMat[M13]
	      && pdMat[M21] == m.pdMat[M21]
	      && pdMat[M22] == m.pdMat[M22]
	      && pdMat[M23] == m.pdMat[M23]
	      && pdMat[M31] == m.pdMat[M31]
	      && pdMat[M32] == m.pdMat[M32]
	      && pdMat[M33] == m.pdMat[M33];
   };
   
   /*
    */
    bool operator != (const Mat3x3T& m) const {
      return pdMat[M11] != m.pdMat[M11]
	      || pdMat[M12] != m.pdMat[M12]
	      || pdMat[M13] != m.pdMat[M13]
	      || pdMat[M21] != m.pdMat[M21]
	      || pdMat[M22] != m.pdMat[M22]
	      || pdMat[M23] != m.pdMat[M23]
	      || pdMat[M31] != m.pdMat[M31]
	      || pdMat[M32] != m.pdMat[M32]
	      || pdMat[M33] != m.pdMat[M33];
   };
   
   /*
    Prodotto matrice per matrice.
    Restituisce il prodotto di se stessa per m in un temporaneo.
    */
   Mat3x3T operator * (const Mat3x3T& m) const
   {
       return Mat3x3T(pdMat[M11]*m.pdMat[M11]+pdMat[M12]*m.pdMat[M21]+pdMat[M13]*m.pdMat[M31],
		 pdMat[M21]*m.pdMat[M11]+pdMat[M22]*m.pdMat[M21]+pdMat[M23]*m.pdMat[M31],
		 pdMat[M31]*m.pdMat[M11]+pdMat[M32]*m.pdMat[M21]+pdMat[M33]*m.pdMat[M31],
		 pdMat[M11]*m.pdMat[M12]+pdMat[M12]*m.pdMat[M22]+pdMat[M13]*m.pdMat[M32],
		 pdMat[M21]*m.pdMat[M12]+pdMat[M22]*m.pdMat[M22]+pdMat[M23]*m.pdMat[M32],
		 pdMat[M31]*m.pdMat[M12]+pdMat[M32]*m.pdMat[M22]+pdMat[M33]*m.pdMat[M32],
		 pdMat[M11]*m.pdMat[M13]+pdMat[M12]*m.pdMat[M23]+pdMat[M13]*m.pdMat[M33],
		 pdMat[M21]*m.pdMat[M13]+pdMat[M22]*m.pdMat[M23]+pdMat[M23]*m.pdMat[M33],
		 pdMat[M31]*m.pdMat[M13]+pdMat[M32]*m.pdMat[M23]+pdMat[M33]*m.pdMat[M33]);
    };      

   T Tr(void) const {
      return pdMat[M11]+pdMat[M22]+pdMat[M33];
   };
      
   /*Input/Output */
   
   /*
    Scrittura su ostream della matrice.
    Scrive se stessa sull'ostream out usando come separatore tra colonne sFill
    e come separatore tra le righe sFill2.
    */
   std::ostream& Write(std::ostream& out, 
		  const char* sFill = " ", 
		  const char* sFill2 = NULL) const;
};

extern const Mat3x3T<doublereal> Eye3;
extern const Mat3x3T<doublereal> Zero3x3;
extern const Vec3T<doublereal> Zero3;

/* Mat3x3T - end */

/*Operazioni esterne su Vec3T e Mat3x3T */

/*
 Operatore "meno" unario su Vec3T.
 Restituisce l'opposto di se stesso in un temporaneo.
 */
template<class T>
extern Vec3T<T> operator - (const Vec3T<T>& v);

/*
 Operatore "meno" unario su Mat3x3T. 
 Restituisce l'opposto di se stesso in un temporaneo.
 */
template<class T>
extern Mat3x3T<T> operator - (const Mat3x3T<T>& v);

/*
 Operatore di scrittura di Vec3T su ostream.
 Nota: i coefficienti sono separati da spazi. Non c'e' endl al termine
 */
template<class T>
extern std::ostream& operator << (std::ostream& out, const Vec3T<T>& v);
      
/*
 Operatore di scrittura di Mat3x3T su ostream.
 Nota: i coefficienti sono separati da spazi e sono scritti consecutivamente,
 per righe. Non c'e' endl al termine.
 */
template<class T>
extern std::ostream& operator << (std::ostream& out, const Mat3x3T<T>& m);


/*
 Funzione di Output di reali su ostream.
 Necessarie per poter generare i templates dei legami costitutivi.
 Il terzo parametro e' definito solo per compatibilita'.
 @param out   ostream su cui avviene la scrittura.
 @param d     valore da scrivere.
 */
template<class T>
extern std::ostream& Write(std::ostream& out, const T& d, const char*);
   
/*
 Funzione di Output di Vec3T su ostream.
 Necessarie per poter generare i templates dei legami costitutivi.
 @param out   ostream su cui avviene la scrittura.
 @param v     vettore da scrivere.
 @param s     separatore tra i valori.
 */
template<class T>
extern std::ostream& Write(std::ostream& out, const Vec3T<T>& v, const char* s = " ");
   
/*
 Funzione di Output di Mat3x3T su ostream.
 Necessarie per poter generare i templates dei legami costitutivi.
 @param out   ostream su cui avviene la scrittura.
 @param d     valore da scrivere.
 @param s     separatore tra i valori (colonne).
 @param s2    separatore tra i valori (righe); se nullo, usa s.
 */
template<class T>
extern std::ostream& Write(std::ostream& out,
		      const Mat3x3T<T>& m,
		      const char* s = " ", 
		      const char* s2 = NULL);


/*
 Calcola i parametri di Rodriguez g a partire dalla matrice di rotazione R.
 Nota: i parametri devono essere definiti, ovvero R non deve rappresentare 
 una rotazione a cui corrispondono parametri singolari.
 */
template<class T>
extern Vec3T<T> gparam(const Mat3x3T<T>& R);
   
/*
 Calcola la matrice di rotazione a partire da due vettori sghembi. 
 @param ia indice del vettore va nel sistema locale.
 @param va vettore ia nel sistema locale.
 @param ib indice del vettore che risulta dall'elaborazione di vb.
 @param vb vettore che viene trasformato come segue:
 il vettore va viene assunto come asse ia nel sistema locale,
 mentre la componente di vb normale a va e giacente nel piano 
 normale al prodotto vettore tra va e vb viene assunta come
 componente ib nel sistema locale.
 */
template<class T>
extern Mat3x3T<T> MatR2vec(unsigned short int ia, 
		       const Vec3T<T>& va, 
		       unsigned short int ib, 
		       const Vec3T<T>& vb);


/*
 Calcola gli angoli di Eulero a partire dalla matrice di rotazione R.
 Nota: gli angoli di Eulero sono ritornati in gradi.
 */
extern const doublereal dRaDegr;
template<class T>
extern Vec3T<T> MatR2EulerAngles(const Mat3x3T<T>& R);
template<class T>
extern void MatR2EulerParams(const Mat3x3T<T>& R, T& e0, Vec3T<T>& e);
template<class T>
extern Vec3T<T> Unwrap(const Vec3T<T>& vPrev, const Vec3T<T>& v);

/*
 Calcola la matrice di rotazione corrispondente agli angoli di Eulero v.
 Nota: gli angoli di Eulero vengono letti in radianti.
 */
template<class T>
extern Mat3x3T<T> EulerAngles2MatR(const Vec3T<T>& v);


/* _MatR_Manip - begin */

// Manipolatore per matrice R con parametri di Rodriguez.
template<class T>
class _MatR_Manip : public _Mat3x3T_Manip<T> {   
 public:
   /*
    Operatore deprecato (e' poco efficiente).
    */
   inline Mat3x3T<T> operator << (const Vec3T<T>& g) const {
      T d = (4./(4.+g.Dot()));
      return Mat3x3T<T>(g*d) + Mat3x3T<T>(g, g*(d/2.)) + Eye3;
   };
   
   /*
    Crea in m la matrice R corrispondente ai parametri g.
    */
   inline void Make(Mat3x3T<T>& m, const Vec3T<T>& g) const {
      T d = (4./(4.+g.Dot()));
      
      /*
       m = Eye3;
       m += Mat3x3T(g*d);
       */
      
      /* E' piu' efficiente se creo contemporaneamente I+d*g/\ */
      m = Mat3x3T<T>(1., g*d);
      
      /* Alla fine sommo il termine d/2*g/\g/\, che e' una matrice piena */
      m += Mat3x3T<T>(g, g*(d/2.));
   };
};

/* _MatR_Manip - end */


/* _MatG_Manip - begin */

// Manipolatore per matrice G con parametri di Rodriguez.
template<class T>
class _MatG_Manip : public _Mat3x3T_Manip<T> {
 public:
   /*
    Operatore deprecato (e' poco efficiente).
    */
   inline Mat3x3T<T> operator << (const Vec3T<T>& g) const {
      T d = (4./(4.+g.Dot()));
      return Mat3x3T<T>(g*(d/2.)) + Eye3 * d;
   };

   /*
    Crea in m la matrice G corrispondente ai parametri g.
    */
   inline void Make(Mat3x3T<T>& m, const Vec3T<T>& g) const {
      T d = (4./(4.+g.Dot()));
      m = Mat3x3T<T>(d, g*(d/2.));
   };
};

/* _MatG_Manip - end */


/* _MatGm1_Manip - begin */

// Manipolatore per inversa della matrice G con parametri di Rodriguez */
template<class T>
class _MatGm1_Manip : public _Mat3x3T_Manip<T> {
 public:
   /*
    Operatore deprecato (e' poco efficiente).
    */
   inline Mat3x3T<T> operator << (const Vec3T<T>& g) const {
      return g.Tens()/4.-Mat3x3T<T>(g/2.) + Eye3;
   };
   
   /*
    Crea in m l'inversa della matrice G corrispondente ai parametri g.
    */
   inline void Make(Mat3x3T<T>& m, const Vec3T<T>& g) const {
      m = Mat3x3T<T>(1., g/(-2.));
      m += g.Tens()/4.;
   };
};

/* _MatGm1_Manip - end */


/*Costanti */
   
/*
 Matrice identita' 3x3 
 */
extern const Mat3x3T<doublereal> Eye3;
   
/*
 Matrice nulla 3x3
 */
extern const Mat3x3T<doublereal> Zero3x3;
   
/*
 Vettore nullo 3
 */
extern const Vec3T<doublereal> Zero3;

/* known orientation descriptions */
enum OrientationDescription {
	UNKNOWN_ORIENTATION_DESCRIPTION	= -1,
	EULER_123			= 0,
	ORIENTATION_VECTOR,
	ORIENTATION_MATRIX,

	LAST_ORIENTATION_DESCRIPTION
};

#endif /* MATVEC3T_H */

