/* 
 * MBDyn (C) is a multibody analysis code. 
 * http://www.mbdyn.org
 *
 * Copyright (C) 1996-2005
 *
 * Pierangelo Masarati  <masarati@aero.polimi.it>
 * Paolo Mantegazza     <mantegazza@aero.polimi.it>
 *
 * Dipartimento di Ingegneria Aerospaziale - Politecnico di Milano
 * via La Masa, 34 - 20156 Milano, Italy
 * http://www.aero.polimi.it
 *
 * Changing this copyright notice is forbidden.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation (version 2 of the License).
 * 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* Copyright Marco Morandini  2005*/

#ifdef HAVE_CONFIG_H
#include "mbconfig.h"           /* This goes first in every *.c,*.cc file */
#endif /* HAVE_CONFIG_H */

#include <iostream>
#include <ac/float.h>

#include "loadable.h"

/*
 * Usage:
 *
 *	loadable: <label>, <module name>, help [ , <module data> ] ;
 */

struct module_wheel3 {
	/*
	 * Ruota
	 */
	StructNode *pWheel;

	/*
	 * Direzione asse ruota
	 */
	Vec3 WheelAxle;
	
	/*
	 * Terreno
	 */
	StructNode *pGround;

	/*
	 * Posizione e orientazione del terreno
	 */
	Vec3 GroundPosition;
	Vec3 GroundDirection;

	/*
	 * Geometria ruota
	 */
	doublereal dRadius;
	doublereal dInternalRadius;
	doublereal dVolCoef;

	doublereal dRefArea;
	doublereal dRNP;		/* R+nG'*pG */
	doublereal dV0;

	/*
	 * Proprieta' pneumatico
	 */
	doublereal dP0;
	doublereal dGamma;
	doublereal dHystVRef;

	/*
	 * Attrito
	 */
	flag fSlip;
	DriveCaller *pMuX0;
	DriveCaller *pMuY0;
	DriveCaller *pMuY1;
	doublereal dvThreshold;

	/*
	 * Output
	 */
	Vec3 F;
	Vec3 M;
	doublereal dInstRadius;
	doublereal dDeltaL;
	doublereal dVn;
	doublereal dSr;
	doublereal dAlpha;
	doublereal dAlphaThreshold;
	doublereal dMuX;
	doublereal dMuY;
	doublereal dVa;
	doublereal dVc;
};

/* funzioni di default */
static void*
read(LoadableElem* pEl,
		DataManager* pDM,
		MBDynParser& HP)
{
	DEBUGCOUTFNAME("read");
	
	/* allocation of user-defined struct */
	module_wheel3* p = NULL;
	SAFENEW(p, module_wheel3);

	/*
	 * help
	 */
	if (HP.IsKeyWord("help")) {
		silent_cout(
"									\n"
"Module: 	wheel3							\n"
"Author: 	Stefania Gualdi <gualdi@aero.polimi.it>			\n"
"		Pierangelo Masarati <masarati@aero.polimi.it>		\n"
"		Marco Morandini <morandini@aero.polimi.it>		\n"
"Organization:	Dipartimento di Ingegneria Aerospaziale			\n"
"		Politecnico di Milano					\n"
"		http://www.aero.polimi.it				\n"
"									\n"
"	All rights reserved						\n"
"									\n"
"Connects 2 structural nodes:						\n"
"     -	Wheel								\n"
"     -	Ground								\n"
"									\n"
"Note: 									\n"
"     -	The Axle and the Wheel structural nodes must be connected 	\n"
"	by a joint that allows relative rotations only about 		\n"
"	one axis (the axle)						\n"
"     -	The center of the wheel is assumed coincident with 		\n"
"	the position of the wheel structural node			\n"
"     -	The Ground structural node supports a plane defined		\n"
"	a point and a direction orthogonal to the plane (future 	\n"
"	versions might use an arbitrary, deformable surface)		\n"
"     -	The forces are applied at the \"contact point\", that 		\n"
"	is defined according to geometrical properties 			\n"
"	of the system and according to the relative position 		\n"
"	and orientation of the Wheel and Ground structural nodes	\n"
"									\n"
"     -	Input:								\n"
"		<wheel structural node label> ,				\n"
"		<wheel axle direction> ,				\n"
"		<ground structural node label> ,			\n"
"		<reference point position of the ground plane> ,	\n"
"		<direction orthogonal to the ground plane> ,		\n"
"		<wheel radius> ,					\n"
"		<torus radius> ,					\n"
"		<volume coefficient (black magic?)> ,			\n"
"		<tire pressure> ,					\n"
"		<tire polytropic exponent> ,				\n"
"		<reference velocity for tire hysteresis>		\n"
"		[ slip ,						\n"
"		<longitudinal friction coefficient drive>		\n"
"		<lateral friction coefficient drive for s.r.=0>		\n"
"		<lateral friction coefficient drive for s.r.=1>		\n"
"		[ , threshold , <slip ratio velocity threshold> , 	\n"
"			<slip angle velocity threshold> ] ]		\n"
"									\n"
"     -	Output:								\n"
"		1)	element label					\n"
"		2-4)	tire force in global reference frame		\n"
"		5-7)	tire couple in global reference frame		\n"
"		8)	effective radius				\n"
"		9)	tire radial deformation				\n"
"		10)	tire radial deformation velocity		\n"
"		11)	slip ratio					\n"
"		12)	slip angle					\n"
"		13)	longitudinal friction coefficient		\n"
"		14)	lateral friction coefficient			\n"
"		15)	axis relative tangential velocity		\n"
"		16)	point of contact relative tangential velocity	\n"
			<< std::endl);

		if (!HP.IsArg()) {
			/*
			 * Exit quietly if nothing else is provided
			 */
			throw NoErr();
		}
	}

	/*
	 * leggo la ruota
	 */
	p->pWheel = (StructNode *)pDM->ReadNode(HP, Node::STRUCTURAL);

	/*
	 * leggo l'orientazione dell'asse ruota nel sistema locale
	 */
	ReferenceFrame RF = ReferenceFrame(p->pWheel);
	p->WheelAxle = HP.GetVecRel(RF);
	
	/*
	 * leggo il terreno
	 */
	p->pGround = (StructNode *)pDM->ReadNode(HP, Node::STRUCTURAL);
	
	/*
	 * leggo posizione ed orientazione del terreno nel sistema del nodo
	 */
	RF = ReferenceFrame(p->pGround);
	p->GroundPosition = HP.GetPosRel(RF);
	p->GroundDirection = HP.GetVecRel(RF);

	/*
	 * normalizzo l'orientazione del terreno
	 */
	doublereal d = p->GroundDirection.Dot();
	if (d <= DBL_EPSILON) {
		silent_cerr("Wheel3(" << pEl->GetLabel() << "): "
			"null direction at line " << HP.GetLineData()
			<< std::endl);
		throw DataManager::ErrGeneric();
	}
	p->GroundDirection /= sqrt(d);

	/*
	 * Geometria ruota
	 */
	p->dRadius = HP.GetReal();
	p->dInternalRadius = HP.GetReal();
	p->dVolCoef = HP.GetReal();
	
	/*
	 * Area di riferimento
	 */
	p->dRefArea = 3.7*p->dInternalRadius*sqrt(
			p->dInternalRadius*(2.*p->dRadius-p->dInternalRadius)
			);

	/*
	 * termine per il calcolo di Delta L
	 */
	p->dRNP = p->dRadius+p->GroundPosition*p->GroundDirection;

	/*
	 * Volume di riferimento
	 */
	p->dV0 = 2.*M_PI*(p->dRadius-p->dInternalRadius)
		*M_PI*p->dInternalRadius*p->dInternalRadius*p->dVolCoef;

	/*
	 * Proprieta' pneumatico
	 */
	p->dP0 = HP.GetReal();
	p->dGamma = HP.GetReal();
	p->dHystVRef = HP.GetReal();
	p->dVn = 0.;

	/*
	 * Attrito
	 */
	p->fSlip = 0;
	if (HP.IsKeyWord("slip")) {
		p->fSlip = 1;

		/*
		 * Parametri di attrito
		 */
		p->pMuX0 = HP.GetDriveCaller();
		p->pMuY0 = HP.GetDriveCaller();
		p->pMuY1 = HP.GetDriveCaller();
	
		p->dvThreshold = 0.;
		p->dAlphaThreshold = 0.;
		if (HP.IsKeyWord("threshold")) {
			p->dvThreshold = HP.GetReal();
			if (p->dvThreshold < 0.) {
				silent_cerr("Wheel3(" << pEl->GetLabel() << "): "
					"illegal velocity threshold " << p->dvThreshold
					<< " at line " << HP.GetLineData() << std::endl);
				p->dvThreshold = fabs(p->dvThreshold);
			}

			p->dAlphaThreshold = HP.GetReal();
			if (p->dAlphaThreshold < 0.) {
				silent_cerr("Wheel3(" << pEl->GetLabel() << "): "
					"illegal slip angle threshold " << p->dAlphaThreshold
					<< " at line " << HP.GetLineData() << std::endl);
				p->dAlphaThreshold = fabs(p->dAlphaThreshold);
			}
		}
	}
	
	std::ostream& out = pDM->GetLogFile();
	out << "wheel3: " << pEl->GetLabel()
		<< " " << p->pWheel->GetLabel()	//node label
		<< " " << p->WheelAxle		//wheel axle
		<< " " << p->pGround->GetLabel()//ground label
		<< " " << p->GroundDirection	//ground direction
		<< " " << p->dRadius		//wheel radius
		<< " " << p->dInternalRadius	//wheel internal radius
		<< std::endl;
	
	return (void *)p;
}

#if 0
static unsigned int
i_get_num_dof(const LoadableElem* pEl)
{
	DEBUGCOUTFNAME("i_get_num_dof");
	return 0;
}
#endif

static void
output(const LoadableElem* pEl, OutputHandler& OH)
{
	DEBUGCOUTFNAME("output");
	
	if (pEl->fToBeOutput()) {
		module_wheel3* p = (module_wheel3 *)pEl->pGetData();      
		std::ostream& out = OH.Loadable();

		out << std::setw(8) << pEl->GetLabel()	/* 1:	label */
			<< " ", p->F.Write(out, " ")	/* 2-4:	force */
			<< " ", p->M.Write(out, " ")	/* 5-7:	moment */
			<< " " << p->dInstRadius	/* 8:	inst. radius */
			<< " " << p->dDeltaL		/* 9:	radial deformation */
			<< " " << p->dVn 		/* 10:	radial deformation velocity */
			<< " " << p->dSr		/* 11:	slip ratio */
			<< " " << 180./M_PI*p->dAlpha	/* 12:	slip angle */
			<< " " << p->dMuX		/* 13:	longitudinal friction coefficient */
			<< " " << p->dMuY		/* 14:	lateral friction coefficient */
			<< " " << p->dVa		/* 15:	axis relative velocity */
			<< " " << p->dVc		/* 16:	POC relative velocity */
			<< std::endl;
	}
}

#if 0
static std::ostream&
restart(const LoadableElem* pEl, std::ostream& out)
{
	DEBUGCOUTFNAME("restart");
	return out << "not implemented yet;" << std::endl;
}
#endif

static void
work_space_dim(const LoadableElem* pEl, 
		    integer* piNumRows, 
		    integer* piNumCols)
{
	DEBUGCOUTFNAME("work_space_dim");
	*piNumRows = 12;
	*piNumCols = 12;
}

#include<vector>
#include "CppAD/CppAD.h"
namespace CppAD {

	AD<doublereal> fabs(const AD<doublereal>& f);
	template<class T>
	AD<doublereal> copysign(T x, AD<doublereal> y) {
		return AD<doublereal>(x*(y/abs(y)));
	};
	AD<doublereal> tanh(const AD<doublereal> &x) {return sinh(x)/cosh(x);};
	doublereal MbdynFakeValue(const doublereal& x) {return x;};
}


/*
Variabili:
- posizione terreno		p->pGround->GetXCurr()
- orientazione ground		p->pGround->GetRCurr()
- posizione ruota		p->pWheel->GetXCurr()
- orientazione ruota		p->pWheel->GetRCurr()
- velocita' ruota		p->pWheel->GetVCurr()
- velocita' terreno		p->pGround->GetVCurr()
- velocita' angolare terreno	p->pGround->GetWCurr()
*/


template<class T>
static void res_vec(module_wheel3* p, std::vector<T>& x_indep, std::vector<T> &y_dep, int label) {

	typedef Vec3T<T> Vec3;
	typedef Mat3x3T<T> Mat3x3;
	
	Vec3 GroundVCurr(x_indep[0],x_indep[1],x_indep[2]);
	Vec3 GroundWCurr(x_indep[3],x_indep[4],x_indep[5]);

	Vec3 GroundXCurr(x_indep[6],x_indep[7],x_indep[8]);
	Vec3 GroundPhi(x_indep[9],x_indep[10],x_indep[11]);
	Mat3x3 GroundRCurr(p->pGround->GetRCurr());
	GroundRCurr += GroundPhi.Cross(GroundRCurr);
	
	Vec3 WheelVCurr(x_indep[12],x_indep[13],x_indep[14]);
	Vec3 WheelWCurr(x_indep[15],x_indep[16],x_indep[17]);

	Vec3 WheelXCurr(x_indep[18],x_indep[19],x_indep[20]);
	Vec3 WheelPhi(x_indep[21],x_indep[22],x_indep[23]);
	Mat3x3 WheelRCurr(p->pWheel->GetRCurr());
	WheelRCurr += WheelPhi.Cross(WheelRCurr);


	/*
	 * Orientazione del terreno nel sistema assoluto
	 */
	
	Vec3 n(GroundRCurr*Vec3(p->GroundDirection));
	
	/*
	 * Distanza Wheel Ground nel sistema assoluto
	 */
	Vec3 x = WheelXCurr-GroundXCurr;

	/*
	 * se dDeltaL > 0 c'e' contatto, altrimenti no
	 */
	T dDeltaL = p->dRNP - x*n;
	T dInstRadius = p->dRadius-dDeltaL;
	
	Vec3 va = WheelVCurr
		-GroundVCurr-(GroundWCurr).Cross(
			GroundRCurr*Vec3(p->GroundPosition)
			);
	p->dVa = CppAD::MbdynFakeValue(((va - n*(n*va)).Norm()));
	
	/*
	 * Velocita' tra Wheel (nel punto di contatto) 
	 * e Ground nel sistema assoluto
	 */
	Vec3 v = va - (WheelWCurr).Cross(n*dInstRadius);
	
	/*
	 * Componente normale al terreno della velocita'
	 * (positiva se la ruota si allontana dal terreno)
	 */
	T dVn = n*v;
	p->dVn = CppAD::MbdynFakeValue(dVn);
	
	p->dVc = CppAD::MbdynFakeValue((v - n*p->dVn).Norm());
	
	/*
	 * Stima dell'area di contatto
	 */
	T dA = p->dRefArea*(dDeltaL/p->dInternalRadius);
	
	/*
	 * Stima del volume di compenetrazione tra ruota e terreno
	 */
	T dDeltaV = .5*dA*dDeltaL;

	/*
	 * Stima della pressione del pneumatico (adiabatica)
	 */
	T dP = p->dP0*pow(p->dV0/(p->dV0 - dDeltaV), T(p->dGamma));

	/*
	 * Il punto di applicazione della forza e' xW - R * nG ;
	 * per la forza normale si puo' usare anche la posizione
	 * della ruota, nell'ipotesi che il punto di contatto
	 * stia nell'intersezione della normale al ground passante
	 * per l'asse ruota.
	 *
	 * FIXME: perche' dRadius invece di dInstRadius?
	 */
	Vec3 pc = WheelXCurr - (n*dInstRadius);

	/*
	 * Forza
	 */
	T dFn = (dA*dP*(1. - tanh(dVn/p->dHystVRef)));
	Vec3 F = n*dFn;
	
	if (p->fSlip) {
	
		/*
		 * Direzione di "avanzamento": asse ruota cross normale
		 * al ground
		 */
		Vec3 fwd = (WheelRCurr*Vec3(p->WheelAxle)).Cross(n);
		T d = fwd.Dot();
		if (d < DBL_EPSILON) {
			silent_cerr("Wheel3(" << label << "): "
				"wheel axle is (neraly) orthogonal "
				"to the ground" << std::endl);
			throw DataManager::ErrGeneric();
		}
		fwd /= sqrt(d);

		/*
		 * Slip ratio
		 */
		T dvx = fwd.Dot(v);
		T sgn = copysign(1., dvx);
		T dfvx = fabs(dvx);
		T dvax = fwd.Dot(va);
		T dfvax = fabs(dvax);

		/*
		 * FIXME: se la vax va a zero (perche' il velivolo si e'
		 * fermato, ad esempio) lo "sleep" ratio deve essere 
		 * "piccolo", o no?
		 */
		T dSr = dfvx/(dfvax + p->dvThreshold);
		if (dSr > 1.) {
			dSr = 1.;
		}
		p->dSr = CppAD::MbdynFakeValue(dSr);

		/*
		 * Direzione laterale: normale cross forward
		 */
		Vec3 lat = n.Cross(fwd);

		/*
		 * Velocita' laterale del mozzo
		 */
		T dvay = lat.Dot(va);

		/*
		 * Angolo di deriva del mozzo
		 * Nota: ristretto al Io-IVo quadrante
		 * questa threshold sul modulo della velocita' fa in modo
		 * che l'angolo vada a zero se il modulo della velocita'
		 * e' troppo piccolo
		 */
		T dAlpha = atan2(dvay, fabs(dvax));
		if (p->dAlphaThreshold > 0.) {
			T dtmp = tanh(sqrt(dvax*dvax + dvay*dvay)/p->dAlphaThreshold);
			dAlpha *= dtmp*dtmp;
		}
		p->dAlpha = CppAD::MbdynFakeValue(dAlpha);
		/*
		 * Coefficiente di attrito longitudinale
		 */
		T dMuX0 = p->pMuX0->dGet(dSr);
		/*
		 * Nota: alpha/(pi/2) e' compreso tra -1. e 1.
		 */
		T dMuX = dMuX0*sgn*(1. -fabs(dAlpha)/M_PI_2);
		p->dMuX = CppAD::MbdynFakeValue(dMuX);
		
		/*
		 * Correggo le forze:
		 * uso il coefficiente di attrito longitudinale
		 * con il segno della velocita' del punto di contatto
		 */
		F -= fwd*dFn*dMuX;

		if (dvay != 0.) {

#if 0 /* non serve piu': Alpha e' ristretta al Io-IVo quadrante */
			if (dAlpha > M_PI_2) {
				dAlpha = M_PI-dAlpha;
			} else if (dAlpha < -M_PI_2) {
				dAlpha = -M_PI-dAlpha;
			}
#endif
			
			T dMuY0 = p->pMuY0->dGet(dAlpha);
			T dMuY1 = p->pMuY1->dGet(dAlpha);
			T dMuY = dMuY0 + (dMuY1 - dMuY0)*dSr;
			p->dMuY = CppAD::MbdynFakeValue(dMuY);
			/*
			 * Correggo le forze
			 */
			F -= lat*dFn*dMuY;
		}
	}

	/*
	 * Momento
	 */
	p->F[0] = CppAD::MbdynFakeValue(F[0]);
	p->F[1] = CppAD::MbdynFakeValue(F[1]);
	p->F[2] = CppAD::MbdynFakeValue(F[2]);

	Vec3 M = (pc - WheelXCurr).Cross(F);
	p->M[0] = CppAD::MbdynFakeValue(M[0]);
	p->M[1] = CppAD::MbdynFakeValue(M[1]);
	p->M[2] = CppAD::MbdynFakeValue(M[2]);
	
	Vec3 Mt = (pc - GroundXCurr).Cross(F);

	y_dep[0]  = -F[0];
	y_dep[1]  = -F[1];
	y_dep[2]  = -F[2];
	y_dep[3]  = -Mt[0];
	y_dep[4]  = -Mt[1];
	y_dep[5]  = -Mt[2];
	y_dep[6]  =  F[0];
	y_dep[7]  =  F[1];
	y_dep[8]  =  F[2];
	y_dep[9]  =  M[0];
	y_dep[10] =  M[1];
	y_dep[11] =  M[2];
}

template void res_vec(module_wheel3* p, std::vector<CppAD::AD<doublereal> >& x_indep, 
		std::vector<CppAD::AD<doublereal> > &y_dep, int label);
template void res_vec(module_wheel3* p, std::vector<doublereal>& x_indep, 
		std::vector<doublereal> &y_dep, int label);

static VariableSubMatrixHandler& 
ass_jac(LoadableElem* pEl, 
	VariableSubMatrixHandler& WorkMat,
	doublereal dCoef, 
	const VectorHandler& XCurr,
	const VectorHandler& XPrimeCurr)
{
	DEBUGCOUTFNAME("ass_jac");   

	module_wheel3* p = (module_wheel3 *)pEl->pGetData();

	integer iGroundFirstMomIndex = p->pGround->iGetFirstMomentumIndex();
	integer iGroundFirstPosIndex = p->pGround->iGetFirstPositionIndex();
	integer iWheelFirstMomIndex = p->pWheel->iGetFirstMomentumIndex();
	integer iWheelFirstPosIndex = p->pWheel->iGetFirstPositionIndex();

	/* Dimensiona la matrice, la azzera e pone gli indici corretti */
	FullSubMatrixHandler& WM = WorkMat.SetFull();   
	WM.ResizeReset(12, 12);
	
	for (int iCnt = 1; iCnt <= 6; iCnt++) {
		WM.PutRowIndex(iCnt, iGroundFirstMomIndex+iCnt);
		WM.PutColIndex(iCnt, iGroundFirstPosIndex+iCnt);
		WM.PutRowIndex(6+iCnt, iWheelFirstMomIndex+iCnt);
		WM.PutColIndex(6+iCnt, iWheelFirstPosIndex+iCnt);
	}

	typedef CppAD::AD<doublereal> T;
	//typedef doublereal T;
	std::vector<T> x_indep(24);
	
	Vec3 vground(p->pGround->GetVCurr());
	x_indep[0] = vground[0];
	x_indep[1] = vground[1];
	x_indep[2] = vground[2];
	Vec3 wground(p->pGround->GetWCurr());
	x_indep[3] = wground[0];
	x_indep[4] = wground[1];
	x_indep[5] = wground[2];
	Vec3 xground(p->pGround->GetXCurr());
	x_indep[6] = xground[0];
	x_indep[7] = xground[1];
	x_indep[8] = xground[2];
	x_indep[9] = 0.;
	x_indep[10] = 0.;
	x_indep[11] = 0.;
	Vec3 vwheel(p->pWheel->GetVCurr());
	x_indep[12] = vwheel[0];
	x_indep[13] = vwheel[1];
	x_indep[14] = vwheel[2];
	Vec3 wwheel(p->pWheel->GetWCurr());
	x_indep[15] = wwheel[0];
	x_indep[16] = wwheel[1];
	x_indep[17] = wwheel[2];
	Vec3 xwheel(p->pWheel->GetXCurr());
	x_indep[18] = xwheel[0];
	x_indep[19] = xwheel[1];
	x_indep[20] = xwheel[2];
	x_indep[21] = 0.;
	x_indep[22] = 0.;
	x_indep[23] = 0.;

	/*
	 * Orientazione del terreno nel sistema assoluto
	 */
	Vec3 n = p->pGround->GetRCurr()*p->GroundDirection;
	
	/*
	 * Distanza Wheel Ground nel sistema assoluto
	 */
	Vec3 x = p->pWheel->GetXCurr()-p->pGround->GetXCurr();

	/*
	 * se dDeltaL > 0 c'e' contatto, altrimenti no
	 */
	doublereal dDeltaL = p->dRNP - x*n;

	/*
	 * Reset dati per output
	 */
	p->dDeltaL = dDeltaL;
	p->dInstRadius = p->dRadius-dDeltaL;
	
	p->dSr = 0.;
	p->dAlpha = 0.;

	p->dMuX = 0.;
	p->dMuY = 0.;

	p->dVa = 0.;
	p->dVc = 0.;

	if (dDeltaL < 0.) {
		
		p->F = Zero3;
		p->M = Zero3;

		p->dInstRadius = p->dRadius;
		p->dDeltaL = 0.;
		
		/*
		 * Non assemblo neppure la matrice ;)
		 */
		WorkMat.SetNullMatrix();

		return WorkMat;
	}
	


	std::vector<T> y_dep(12);
	
	CppAD::Independent(x_indep);
	res_vec(p, x_indep, y_dep, pEl->GetLabel());
	CppAD::ADFun<doublereal> f(x_indep, y_dep);

		
	std::vector<doublereal> xx(24);
	for (int i=0; i<24; i++) {
		xx[i] = CppAD::Value(x_indep[i]);
	}
	std::vector<doublereal> J(12*24);
	J = f.Jacobian(xx);
	
	for (integer row = 0; row < 12; row++) {
		for (integer col = 0; col < 6; col++) {
			WM.PutCoef(row+1, col+1, -J[row*24 + col]);
			WM.IncCoef(row+1, col+1, -J[row*24 + col + 6] * dCoef);
			WM.PutCoef(row+1, col+1 + 6, -J[row*24 + col + 12]);
			WM.IncCoef(row+1, col+1 + 6, -J[row*24 + col + 12 + 6] * dCoef);
		}
	}

	
	return WorkMat;
}

static SubVectorHandler& 
ass_res(LoadableElem* pEl, 
	SubVectorHandler& WorkVec,
	doublereal dCoef,
	const VectorHandler& XCurr, 
	const VectorHandler& XPrimeCurr)
{
	DEBUGCOUTFNAME("ass_res");
	
	//typedef CppAD::AD<doublereal> T;
	typedef doublereal T;
	std::vector<T> x_indep(24);
	
	module_wheel3* p = (module_wheel3 *)pEl->pGetData();

	Vec3 vground(p->pGround->GetVCurr());
	x_indep[0] = vground[0];
	x_indep[1] = vground[1];
	x_indep[2] = vground[2];
	Vec3 wground(p->pGround->GetWCurr());
	x_indep[3] = wground[0];
	x_indep[4] = wground[1];
	x_indep[5] = wground[2];
	Vec3 xground(p->pGround->GetXCurr());
	x_indep[6] = xground[0];
	x_indep[7] = xground[1];
	x_indep[8] = xground[2];
	x_indep[9] = 0.;
	x_indep[10] = 0.;
	x_indep[11] = 0.;
	Vec3 vwheel(p->pWheel->GetVCurr());
	x_indep[12] = vwheel[0];
	x_indep[13] = vwheel[1];
	x_indep[14] = vwheel[2];
	Vec3 wwheel(p->pWheel->GetWCurr());
	x_indep[15] = wwheel[0];
	x_indep[16] = wwheel[1];
	x_indep[17] = wwheel[2];
	Vec3 xwheel(p->pWheel->GetXCurr());
	x_indep[18] = xwheel[0];
	x_indep[19] = xwheel[1];
	x_indep[20] = xwheel[2];
	x_indep[21] = 0.;
	x_indep[22] = 0.;
	x_indep[23] = 0.;

	/*
	 * Orientazione del terreno nel sistema assoluto
	 */
	Vec3 n = p->pGround->GetRCurr()*p->GroundDirection;
	
	/*
	 * Distanza Wheel Ground nel sistema assoluto
	 */
	Vec3 x = p->pWheel->GetXCurr()-p->pGround->GetXCurr();

	/*
	 * se dDeltaL > 0 c'e' contatto, altrimenti no
	 */
	doublereal dDeltaL = p->dRNP - x*n;

	/*
	 * Reset dati per output
	 */
	p->dDeltaL = dDeltaL;
	p->dInstRadius = p->dRadius-dDeltaL;
	
	p->dSr = 0.;
	p->dAlpha = 0.;

	p->dMuX = 0.;
	p->dMuY = 0.;

	p->dVa = 0.;
	p->dVc = 0.;

	if (dDeltaL < 0.) {
		
		p->F = Zero3;
		p->M = Zero3;

		p->dInstRadius = p->dRadius;
		p->dDeltaL = 0.;
		
		/*
		 * Non assemblo neppure il vettore ;)
		 */
		WorkVec.Resize(0);

		return WorkVec;
	}
	
	/*
	 * Dimensiono il vettore
	 */
	integer iNumRows = 0;
	integer iNumCols = 0;
	pEl->WorkSpaceDim(&iNumRows, &iNumCols);
   
	WorkVec.ResizeReset(iNumRows);

	integer iGroundFirstMomIndex = p->pGround->iGetFirstMomentumIndex();
	integer iWheelFirstMomIndex = p->pWheel->iGetFirstMomentumIndex();

	/*
	 * Indici equazioni
	 */
	for (int iCnt = 1; iCnt <= 6; iCnt++) {
		WorkVec.PutRowIndex(iCnt, iGroundFirstMomIndex+iCnt);
		WorkVec.PutRowIndex(6+iCnt, iWheelFirstMomIndex+iCnt);
	}

	std::vector<T> y_dep(12);
	res_vec(p, x_indep, y_dep, pEl->GetLabel());
	for (int i=0; i<12; i++) {
		WorkVec.Put(i+1, CppAD::MbdynFakeValue(y_dep[i]));
	}
	return WorkVec;
}

#if 0
static void
before_predict(const LoadableElem* pEl, 
		VectorHandler& X,
		VectorHandler& XP,
		VectorHandler& XPrev,
		VectorHandler& XPPrev)
{
   DEBUGCOUTFNAME("before_predict");
}

static void
after_predict(const LoadableElem* pEl, 
		VectorHandler& X,
		VectorHandler& XP)
{
	DEBUGCOUTFNAME("after_predict");
}

static void
update(LoadableElem* pEl, 
		const VectorHandler& X,
		const VectorHandler& XP)
{
	DEBUGCOUTFNAME("update");
}

static unsigned int
i_get_initial_num_dof(const LoadableElem* pEl)
{
	DEBUGCOUTFNAME("i_get_initial_num_dof");
	return 0;
}

static void
initial_work_space_dim(const LoadableElem* pEl, 
		integer* piNumRows, 
		integer* piNumCols)
{
	DEBUGCOUTFNAME("initial_work_space_dim");
	*piNumRows = 0;
	*piNumCols = 0;
}

static VariableSubMatrixHandler& 
initial_ass_jac(LoadableElem* pEl, 
		VariableSubMatrixHandler& WorkMat, 
		const VectorHandler& XCurr)
{
	DEBUGCOUTFNAME("initial_ass_jac");
	integer iNumRows = 0;
	integer iNumCols = 0;
	pEl->InitialWorkSpaceDim(&iNumRows, &iNumCols);
   
	FullSubMatrixHandler& WM = WorkMat.SetFull();
	WM.ResizeReset(iNumRows, iNumCols);
#if 0	
	module_wheel* p = (module_wheel *)pEl->pGetData();
#endif /* 0 */
   
	/* set sub-matrix indices and coefs */

	return WorkMat;
}

static SubVectorHandler& 
initial_ass_res(LoadableElem* pEl, 
		SubVectorHandler& WorkVec, 
		const VectorHandler& XCurr)
{
	DEBUGCOUTFNAME("initial_ass_res");
	integer iNumRows = 0;
	integer iNumCols = 0;
	pEl->WorkSpaceDim(&iNumRows, &iNumCols);
	
	WorkVec.Resize(iNumRows);

#if 0
	module_wheel* p = (module_wheel *)pEl->pGetData(); 
#endif /* 0 */
	
	/* set sub-vector indices and coefs */
   
	return WorkVec;
}

static void
set_value(const LoadableElem* pEl, DataManager *pDM,
		VectorHandler& X, VectorHandler& XP,
		SimulationEntity::Hints *ph)
{
	DEBUGCOUTFNAME("set_value");
}
   
static void
set_initial_value(const LoadableElem* pEl, VectorHandler& X)
{
	DEBUGCOUTFNAME("set_initial_value");
}
#endif

static unsigned int
i_get_num_priv_data(const LoadableElem* pEl)
{
	DEBUGCOUTFNAME("i_get_num_priv_data");
	return 0;
}

static void
destroy(LoadableElem* pEl)
{
	DEBUGCOUTFNAME("destroy");
	module_wheel3* p = (module_wheel3 *)pEl->pGetData();
	
	SAFEDELETE(p);
}

static int
i_get_num_connected_nodes(const LoadableElem* /* pEl */ )
{
	DEBUGCOUTFNAME("i_get_num_connected_nodes");
	
	/* wheel + ground */
	return 2;
}

static void
get_connected_nodes(const LoadableElem* pEl, 
		int& NumNodes, Node::Type* NdTyp, unsigned int* NdLabels)
{
	DEBUGCOUTFNAME("get_connected_nodes");
	module_wheel3* p = (module_wheel3 *)pEl->pGetData();
	
	/* wheel + ground */
	NumNodes = 2;

	NdTyp[0] = Node::STRUCTURAL;
	NdTyp[1] = Node::STRUCTURAL;

	NdLabels[0] = p->pWheel->GetLabel();
	NdLabels[1] = p->pGround->GetLabel();
}

#ifdef MBDYN_MODULE
static
#endif /* MBDYN_MODULE */
struct
LoadableCalls module_wheel3_lc = {
	LOADABLE_VERSION_SET(1, 4, 0),

	"wheel3",
	"1.2.0",
	"Dipartimento di Ingegneria Aerospaziale, Politecnico di Milano",
	"tire force model for landing gear analysis\n"
		"\tcontact Stefania Gualdi <gualdi@aero.polimi.it>",

	read, /* */
	NULL /* i_get_num_dof */ ,
	NULL /* set_dof */ ,
	output, /* */
	NULL /* restart */ ,
	work_space_dim, /* */
	ass_jac, /* */
	NULL /* ass_mats */ ,
	ass_res, /* */
	NULL /* before_predict */ ,
	NULL /* after_predict */ ,
	NULL /* update */ ,
	NULL /* after_convergence */ ,
	NULL /* i_get_initial_num_dof */ ,
	NULL /* initial_work_space_dim */ ,
	NULL /* initial_ass_jac */ ,
	NULL /* initial_ass_res */ ,
	NULL /* set_value */ ,
	NULL /* set_initial_value */ ,
	i_get_num_priv_data,
	NULL /* i_get_priv_data_idx */ ,
	NULL /* d_get_priv_data */ ,
	i_get_num_connected_nodes,
	get_connected_nodes,
	destroy,
	NULL,
	NULL,
	NULL,
	NULL
};

extern "C" {

/*void *callwheel3 = &module_wheel3_lc;*/
void *wheel3 = &module_wheel3_lc;

#ifndef STATIC_MODULES
extern "C" int
module_init(const char *s, void *dm, void *)
{
	DataManager *pDM = (DataManager *)dm;
	
	pDM->SetLoadableElemModule(module_wheel3_lc.name, &module_wheel3_lc);

	return 0;
}
#endif /* !STATIC_MODULES */

}

